import os
import time
import json
import hashlib
from collections import defaultdict

from langchain import LLMChain
from langchain.llms import OpenAI
from langchain.embeddings.openai import OpenAIEmbeddings
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.vectorstores import Chroma
from langchain.vectorstores import FAISS
from langchain.docstore.document import Document
from langchain import PromptTemplate
from langchain.indexes.vectorstore import VectorstoreIndexCreator
from langchain.chains.question_answering import load_qa_chain
from langchain.memory import ConversationTokenBufferMemory, ConversationKGMemory
from langchain.document_loaders import UnstructuredHTMLLoader
from langchain.document_loaders import DirectoryLoader
from langchain.document_loaders import PyPDFLoader
from langchain.document_loaders import PyMuPDFLoader
from langchain.document_loaders import PDFMinerLoader
# from langchain.document_loaders import PyPDFium2Loader

import chromadb
from chromadb.config import Settings

from . import agent_events

__all__ = ['LLMDocGenerator']

class LLMDocGenerator(object):
    """
    Class to generate text documents using LLMs
    A set of documents that act as context
    against which the agent has to generate text
    """

    def __init__(self,
                 name,
                 cred={},
                 model="openai",
                 searchapi="serpapi"):
        """
        init the LLM query agent
        name: name of the agent
        cred: credentials object
        model: name of the model backend to use
                default to OpenAI GPT model for now
                will be extended in the future to suuport other models
        searchapi: name of the search API backend to use
                    default to serpapi for now
        """

        start_time = time.time()

        # defaults
        self.chunk_size = 1000
        self.chunk_overlap = 100
        self.latest_context = []
        self.context_topK = 1
        self.current_kg = []
        self.metadata = {}
        self.index = None
        self.vdb_client = None
        self.index_name = None
        self.store = None

        # name
        self.agent_name = name

        # creds
        self.cred = cred
        # LLM params
        self.model = model
        self.searchapi = searchapi
        self.chaintype = "stuff"

        # init the llm and embeddings objects
        self.llm, self.embeddings = self._get_llm_objs(model=self.model,
                                                       cred=self.cred)

        # init the QnA chain for internal queries
        prompt = self._get_query_prompt_answer()
        self.llm_chain = load_qa_chain(self.llm,
                                       chain_type=self.chaintype,
                                       prompt=prompt)

        # note metadata for this agent
        self.metadata = {
            "agent": {
                "name": self.agent_name,
                "model": self.model,
                "searchapi": self.searchapi,
                "chaintype": self.chaintype,
            },
            "events": []
        }
        # log that the agent is ready
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_READY, duration)


    ## helper functions

    def _create_id(self, string):
        """
        create a unique identifier from a string
        """
        return hashlib.md5(string.encode('utf-8')).hexdigest()

    def _log_event(self, event_name, duration=0, params={}):
        """
        format an event object to store in the agent's metadata
        event_name: name of the event to log
        duration: how long did this event take, only needed for events
                   that require some waiting (e.g. RTT to LLM APIs)
        params: any other info needed to be logged
        """

        ts = time.time() - duration # to get the actual start time
        event_id = self._create_id(f'{event_name}-{ts}')

        event = {
            "id": event_id,
            "agent": self.agent_name,
            "timestamp": round(ts, 3),
            "duration": round(duration, 3),
            "name": event_name,
            "params": params,
        }
        self.metadata['events'].append(event)

        return event

    def _get_llm_objs(self, model, cred):
        # get the api key from creds
        api_key = self._get_api_key(cred, model)

        # init the model
        if model == "openai":
            # get the llm object
            llm = OpenAI(temperature=0,
                         max_tokens=512,
                         openai_api_key=api_key,
                         request_timeout=60)
            # get the embeddings object
            embeddings = OpenAIEmbeddings(openai_api_key=api_key)
        else:
            llm = None
            embeddings = None

        return llm, embeddings

    def _get_query_prompt_answer(self):
        """
        generate a prompt for running a query in internal mode
        """
        template = """You are an assistant to a prolific venture capilatist assessing new startups for feasibility of funding.
You will be given a number of context documents about a startup.
Your task is to use the information in the context documents to answer questions about the startup.
Respond in the manner of writing an investor memo.
If you can't find the answer in the context documents, DO NOT MAKE UP AN ANSWER.

        Here are the context documents:
        {context}

        Question: {question}
        Answer:
        """

        prompt = PromptTemplate(
            input_variables=["question", "context"],
            template=template
        )

        return prompt

    def _get_api_key(self, cred, key):
        """
        get the API key from the cred
        """
        api_key = None
        if isinstance(cred, str):
            api_key = cred
        if isinstance(cred, dict) and key in cred:
            api_key = cred[key]
        return api_key

    def _err_msg(self, t):
        msgs = {
            "field": "I'm having trouble understanding. Try another way of wording your query."
        }
        return msgs.get(t, "Something went wrong, try your query again.")

    def _add_docset_chroma(self, index, data):
        """
        populate the index
        """
        ids = [doc.metadata.get("id", self._create_id(doc.page_content)) for doc in data]
        docs = [doc.page_content for doc in data]
        metas = [doc.metadata for doc in data]
        index.add(
            documents=docs,
            metadatas=metas,
            ids=ids)
        return

    def _delete_docset_chroma(self, index, data):
        """
        delete from the index where source matches incoming data
        """
        sources = list(set([doc.metadata.get('source') for doc in data]))
        if len(sources) == 1:
            where_clause = {"source": sources[0]}
        else:
            where_clause = {
                "$or": [{"source": source} for source in sources]
            }
        index.delete(
            where=where_clause
        )
        return


    ## interfaces

    def validate_spec(self, spec):

        if isinstance(spec, dict):
            reqd_keys = ['sections']
            missing = [k for k in reqd_keys if k not in spec]
            if len(missing)>0:
                raise Exception("Invalid specification")
        return

    def load_spec(self, profilespec):
        """
        load the spec for generating the document
        profilespec: path to spec file
        """
        # check path
        if isinstance(profilespec, dict):
            self.validate_spec(profilespec)
            return profilespec

        if os.path.exists(profilespec) and profilespec.endswith(".json"):
            with open(profilespec, "r") as fd:
                spec = json.load(fd)
                self.validate_spec(spec)
                return spec

        return None


    def get_metadata(self):
        """
        return metadata collected by the agent
        """
        return self.metadata

    def chunk_data(self, data):
        """
        create chunks from the data
        and add any needed metadata
        """

        def cleanup_metadata(data):
            # take in a list of data document objects
            # and clean up metadata
            curr_source = None
            for i in range(0, len(data)):
                source = data[i].metadata['source']
                data[i].metadata['file'] = source.split('/')[-1]
                if curr_source != source:
                    curr_source = source
                    chunk = 1
                data[i].metadata['chunk'] = chunk
                data[i].metadata['id'] = self._create_id(f"{source}-{chunk}")
                chunk += 1

                ##
                ## add any other custom metadata here
                ##

            return data

        # chunk the data
        text_splitter = RecursiveCharacterTextSplitter(chunk_size=self.chunk_size,
                                                       chunk_overlap=self.chunk_overlap)
        chunks = text_splitter.split_documents(data)

        # add metadata
        chunks = cleanup_metadata(chunks)

        return chunks

    def load_data(self, source, content, metadata={}, params={}):
        """
        set the datasource loader, loads and cleans the data
        source:
            'dir' points to a folder with one or more files to load
            'pdf' points to a single PDF file to load
            'str' contains a text string to load
        content: data content (path, text) depending on type of source
        params: extra params needed for specific loaders
                    glob: what glob filter to use if source=='dir'
                    pdfloader: what type of pdf loader module to use if source=='pdf'
        metadata: any custom metadata to pass when source=='str'
        """

        start_time = time.time()

        if source == 'dir':
            glob = params.get("glob", "**/*.*")
            loader = DirectoryLoader(content, glob=glob, recursive=True)
            data = loader.load()

        elif source == 'pdf':
            pdfloader = params.get("pdfloader", "pymupdf")
            if pdfloader == "pymupdf":
                loader = PyMuPDFLoader(content)
                data = loader.load()
            elif pdfloader == "pypdf":
                loader = PyPDFLoader(content)
                data = loader.load_and_split()
            elif pdfloader == "pypdfium2":
                loader = PyPDFium2Loader(content)
                data = loader.load()
            elif pdfloader == "pdfminer":
                loader = PDFMinerLoader(content)
                data = loader.load()
            else:
                data = None

        elif source == 'str':
            # special handling for string inputs
            metadata["source"] = source
            data = [Document(page_content=content, metadata=metadata)]

        else:
            data = None

        # chunk the data
        data = self.chunk_data(data)

        # log that the data loader is ready
        duration = time.time() - start_time
        params = {
            "source": source,
            "content": content,
            "params": params,
            "metadata": metadata,
        }
        event = self._log_event(agent_events._EVNT_DATA, duration, params=params)

        return data

    def get_index(self):
        """
        return the index object
        """
        return self.index

    def create_add_index_chroma(self, data, persist_directory=None):
        """
        Init the chromadb index and populate it with a set of documents
        """
        # init the ChromaDB client
        self.vdb_client = chromadb.Client(Settings(
            chroma_db_impl="duckdb+parquet",
            persist_directory=persist_directory # Optional, defaults to .chromadb/ in the current directory
        ))

        # make sure we're starting with a fresh db and collection
        self.vdb_client.reset()
        index = self.vdb_client.get_or_create_collection(name=self.index_name,
                                                         embedding_function=self.embeddings.embed_documents)

        # populate the collection
        self._add_docset_chroma(index, data)

        # persist the index to disk
        if persist_directory:
            self.vdb_client.persist()

        return index

    def add_to_index_chroma(self, data):
        """
        add document(s) to a chromadb index
        """
        # first, delete all existing docs from the same sources
        # as what we are adding, we don't want duplicates
        self._delete_docset_chroma(self.index, data)

        # now, add the new docs
        self._add_docset_chroma(self.index, data)

        # persist the db to disk
        self.vdb_client.persist()

        return

    def add_to_index(self, data):
        """
        add document(s) to the agent's index
        """
        start_time = time.time()

        if self.index:
            if self.store == 'chroma':
                self.add_to_index_chroma(data)
            else:
                raise Exception(f"{self.store} does not support adding document")
        else:
            raise Exception("No available index, cannot add document")

        # log that the doc is added
        if self.index:
            duration = time.time() - start_time
            params = {
                "n_items": len(data),
            }
            event = self._log_event(agent_events._EVNT_INDEXADD, duration, params=params)

        return

    def create_add_index(self, data, store='chroma', persist_directory=None, index_name=None):
        """
        create an index from a data source
        data: list of langchain Document() objects
        store: type of vectorstore to use (chroma, faiss, ...)
        """

        start_time = time.time()

        # note what store we are using and the index name
        self.store = store
        self.index_name = self.agent_name if not index_name else index_name

        # create the index
        if store == 'faiss':
            self.index = FAISS.from_documents(data, self.embeddings)
        elif store == 'chroma':
            self.index = self.create_add_index_chroma(data, persist_directory=persist_directory)
        else:
            self.index = None
            self.store = None

        # log that the index is ready
        if self.index:
            duration = time.time() - start_time
            params = {
                "store": store,
                "persist_directory": persist_directory,
                "index_name": self.index_name,
                "n_items": len(data)
            }
            event = self._log_event(agent_events._EVNT_INDEXCREATE, duration, params=params)

        return

    def load_index(self, persist_directory, index_name, store='chroma'):
        """
        load an already persisted index from a directory
        persist_directory: location of persisted index
        store: type of vectorstore to use (chroma, ...)
                only supports chroma for now
        """
        start_time = time.time()

        # make note of the store type
        self.store = store

        # load the index
        if self.store == 'chroma':
            self.vdb_client = chromadb.Client(Settings(
                                    chroma_db_impl="duckdb+parquet",
                                    persist_directory=persist_directory
                                ))
            self.index_name = index_name
            index = self.vdb_client.get_collection(name=self.index_name,
                                                  embedding_function=self.embeddings.embed_documents)
        else:
            index = None

        # log that the index is ready
        duration = time.time() - start_time
        params = {
            "store": self.store,
            "persist_directory": persist_directory,
        }
        event = self._log_event(agent_events._EVNT_INDEXLOAD, duration, params=params)

        self.index = index

        return

    def get_index_stats(self):
        """
        return some stats about the agent's index
        """
        stats = None
        if self.index:
            try:
                stats = {
                    "name": self.index_name,
                    "store": self.store,
                    "n_items": self.index.count()
                }
            except:
                raise Exception("Index does not support stats")
        return stats

    def search_chromadb(self, query, k=7, include_metadata=False):
        """
        run a search against the chromadb index for a list of queries
        """
        # perform query
        results = self.index.query(
                    query_texts=[query],
                    n_results=k,
                    where=None,
                    where_document=None)
        # construct result docset
        docs = []
        for i in range(0, len(results['documents'][0])):
            page_content = results['documents'][0][i]
            metadata = results['metadatas'][0][i]
            doc = Document(page_content=page_content, metadata=metadata)
            docs.append(doc)

        return docs

    def get_similar_docs(self, query, topk=7):
        """
        get top-K similar docs from an index given a query
        query: query string
        index: index object
        topk: number of top-K similar docs to matching query to return
        """
        if self.index:
            if self.store == 'faiss':
                docs = self.index.similarity_search(query,
                                                    k=topk,
                                                    include_metadata=True)
            elif self.store == 'chroma':
                docs = self.search_chromadb(query,
                                            k=topk,
                                            include_metadata=True)
            else:
                docs = None

        return docs

    def generate_query_prompt_kwords(self, context=""):
        """
        generate a prompt for extracting keywords from a paragraph
        """
        template = """You are the chief of staff to a busy executive. You will be given a paragraph of text
        and must identify at most five key phrases from the paragraph. Do not summarize what you find,
        respond with the key phrases exactly. Structure your response as a json list exactly.
        DO NOT RESPOND WITH ANY KEY PHRASES FROM CONTENT ABOVE THIS LINE.

        Here is the paragraph:
        {context}

        The key phrases are:
        """

        prompt = PromptTemplate(
            input_variables=["context"],
            template=template,
        )

        return prompt.format(context=context)

    def run_query_kwords(self, context=""):
        """
        run a query using llm on an internal docset indexed in index
        this is useful when looking for answers that generic llm can provide
        """
        # augment the query with some context to guide the LLM
        query = self.generate_query_prompt_kwords(context)
        result = self.llm(query)
        result = result.strip()
        # a few tries to extract the response
        # sometimes, the LLM messes up
        try:
            result = json.loads(result)
        except:
            try:
                result = json.loads(f"[{result.split('[')[-1]}")
            except:
                pass

        return result

    def generate_query_prompt_collation(self, text, intent, method):
        """
        generate a prompt for summarizing text
        """

        if method == 'summarize':
            template = """Given the following long text (given in backquotes below), summarize it to be {intent}.

            ```
            {text}
            ```

            Summary:
            """
        elif method == 'rewrite':
            template = """Rewrite the following text (given in backquotes below), {intent}, to make it more coherent.

            ```
            {text}
            ```

            Response:
            """
        else:
            template = None

        prompt = PromptTemplate(
            input_variables=["intent", "text"],
            template=template,
        )

        return prompt.format(intent=intent, text=text)

    def run_query_collation(self, intent, text, method):
        """
        run a query using llm to summarize multiple paras of text
        """
        # augment the query with some context to guide the LLM
        prompt = self.generate_query_prompt_collation(text, intent, method)
        result = self.llm(prompt)
        result = result.strip()
        result = {
            "intent": intent,
            "text": text,
            "answer": result,
            "sources": [{"content": text, "source": f"doc-collation-{method}"}]
        }

        return result

    def run_query_answer(self, query):
        """
        run a query using llm on an internal docset indexed in index
        this is useful when looking for answers using a private source of data
        """
        # get the similar docs
        docs = self.get_similar_docs(query)

        # setup the QnA chain object
        response = self.llm_chain({"input_documents":docs, "question":query},
                                    return_only_outputs=True)

        # run the query against the similar docs
        result = {
            "question": query,
            "answer": response.get('output_text', self._err_msg('field')).strip(),
            "sources": [{"content": d.page_content, "source": d.metadata['source']} for d in docs]
        }

        # check if suggest call is needed
        if ('output_text' not in response) or ("i am not sure" in result['answer'].lower()):
            # we don't have a usable answer, so no need for sources
            result['sources'] = []

        return result

    def prompt(self, prompt, mode="answer"):
        """
        run a query on an index using an llm chain object
        prompt: prompt dict containing
            query: when mode=answer
            intent: when mode=summarize
            text: when mode=summarize
        index: index object
        llm: llm object
        mode: 'internal' for querying over docset,
        context: text used to guide the LLM when running in 'external' mode
        """

        start_time = time.time()

        result = None
        if mode == 'answer':
            query = prompt.get('query')
            if query:
                result = self.run_query_answer(query)
        elif mode == 'collate':
            text = prompt.get('text')
            intent = prompt.get('intent', "one para")
            collation = prompt.get('collate', "summarize")
            if text:
                result = self.run_query_collation(intent=intent, text=text, method=collation)
        else:
            pass

        if result:
            answer = result['answer']
            # add keywords identified to the result
            result['keywords'] = self.run_query_kwords(context=answer)

        # log the event
        params = {
            "prompt": prompt,
            "mode": mode,
            "result": result.copy(),
        }
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_QUERY, duration, params=params)

        # add the event to the result
        result['metadata'] = {
            "timestamp": event['timestamp'],
            "duration": event['duration'],
        }

        return result

    def generate_doc(self, profilespec="profilespec.json"):
        """
        generate the document
        """

        start_time = time.time()

        # load the spec first
        self.spec = self.load_spec(profilespec)
        if self.spec is None:
            return None

        # get prefix, suffix
        prefix = self.spec.get('prompt', {}).get('prefix', "")
        suffix = self.spec.get('prompt', {}).get('suffix', "")

        # get sections
        sections = self.spec.get('sections')

        full_text = "-- BEGIN: Investor Memo --" + "\n\n\n"

        for section, detail in sections.items():

            enable = detail.get('enable', True)
            if not enable:
                continue

            # form the section header
            header = detail['title']
            header_text = f'{header}' + "\n" + '_'*len(header) + "\n\n"

            # run through each question for that section
            section_text = ""
            section_sources = []
            section_keywords = []
            for query in detail['questions']:
                # query using the agent
                query = f"{prefix} {query} {suffix}"
                prompt = { "query": query }
                result = self.prompt(prompt, mode="answer")
                answer = result['answer']
                sources = result['sources']
                keywords = result['keywords']
                section_sources += [source['source'] for source in sources]
                section_keywords += keywords

                # form the answer for this question
                section_text += f'{answer}' + "\n\n"

            # summarize the section if needed
            collate = detail.get('collate')
            if collate:
                intent = detail.get('intent', detail.get('title'))
                prompt = {
                    "text": section_text,
                    "intent": intent,
                    "collate": collate,
                }
                result = self.prompt(prompt, mode="collate")
                section_text = result['answer']
                section_text += "\n\n"
                section_keywords = result['keywords']

            # add the keywords for this section
            section_keywords = list(set(section_keywords))
            section_text += "Keywords: " + ", ".join(section_keywords) + "\n\n"

            # add the sources for this section
            section_sources = list(set(section_sources))
            section_text += "Sources: " + ", ".join(section_sources) + "\n\n\n"

            # create the section text for this section
            full_text += f'{header_text}{section_text}'

        full_text += "\n-- Investor Memo :END --"

        # log the event
        params = {
            "spec": self.spec
        }
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_DOCGEN, duration, params=params)

        return full_text

def get_sample_profilespec():

    return {
	    "prompt": {
		    "prefix": "write a para about",
		    "suffix": "for this startup company"
	    },
	    "sections": {
		    "about": {
			    "enable": True,
			    "title": "About the Company",
			    "intent": "introducing the startup",
			    "questions": [
				    "industry and specific focus",
				    "key technologies developed",
				    "unique value proposition"
			    ],
			    "collate": False,
		    },
		    "founders": {
			    "enable": True,
			    "title": "Company Founders",
			    "intent": "about the co-founders of the startup",
			    "questions": [
				    "co-founders and the roles they hold",
				    "past experience of co-founders"
			    ],
			    "collate": "rewrite"
		    },
		    "market-analysis": {
			    "enable": True,
			    "title": "Market Analysis",
			    "intent": "about market analysis for this startup",
			    "questions": [
				    "target customer base",
				    "other potential customers",
				    "other comparable companies and startups in this space",
				    "the main competitors",
				    "biggest risk factors",
				    "key differentiators",
				    "market trends for the domain"
			    ],
			    "collate": False
		    }
	    }
    }

if __name__ == "__main__":

    # vars
    dirpath = os.path.expandvars("$DATA_ROOT/$AGENTNAME/acme_1/bulk")
    singlepdf = os.path.expandvars("$DATA_ROOT/$AGENTNAME/acme_1/single/spotlight.pdf")
    persist_directory = os.path.expandvars("$DATA_ROOT/$AGENTNAME/acme_1/chromadb_index")
    index_name = "acme_1_index"
    profilespec = get_sample_profilespec()
    outfile = None # /path/to/generated/document

    #cred = get_credentials_by_name('openai-api')
    cred = None # environment variable...

    # we'll first create one agent
    # point it to a folder with multiple files to create an index
    # then add one more file to the index in a separate step

    # create the docgen agent
    agent = LLMDocGenerator(name="agent_acme_1",
                            cred=cred)

    # point it to the folder containing multiple pdfs
    data = agent.load_data(source="dir",
                            content=dirpath,
                            params={"glob":"**/*.pdf"})
    # create the index atop the data
    agent.create_add_index(data=data,
                            store="chroma",
                            persist_directory=persist_directory,
                            index_name=index_name)
    # check index stats
    print (json.dumps(agent.get_index_stats(), indent=4))

    # now, add one more pdf using the PyMuPDF loader
    data = agent.load_data(source="pdf",
                            content=singlepdf,
                            params={"pdfloader":"pymupdf"})
    # check index stats again
    print (json.dumps(agent.get_index_stats(), indent=4))
    # we now have the index populated with all the data we need

    # we can run the document generation step with this agent
    # or we can create another agent, point it to the index we created
    # and use that agent to generate the document
    # let's do the latter

    # create a new docgen agent
    agent2 = LLMDocGenerator(name="agent_acme_2")
    # point it to the index we created earlier
    # use the same index name
    agent2.load_index(persist_directory=persist_directory,
                        index_name=index_name,
                        store="chroma")
    # check index stats, it should match what the first agent has
    print (json.dumps(agent2.get_index_stats(), indent=4))

    # use the second agent to generate the document and save
    profilespec = get_sample_profilespec()
    full_text = agent2.generate_doc(profilespec=profilespec)
    if outfile is None:
        print(full_text)
    else:
        with open(outfile, "w") as fd:
            fd.write(full_text)
        print(f"Output in {outfile}")



#!/usr/bin/env python3
"""

Script to extract tables from pdfs or image files. This will be piped through the textract

"""
import os
import sys
import json
import tempfile
import traceback
import re
import inspect
import importlib

import click

from asgiref.sync import async_to_sync, sync_to_async

from langchain.vectorstores import Chroma
from langchain.embeddings.openai import OpenAIEmbeddings

import llmsdk
from llmsdk.lib import *
from llmsdk.agents import *
from llmsdk.services.lib import get_files, get_agent_details
from llmsdk.services.qna import get_qna_agent, supported_exts as qna_supported_exts
from llmsdk.services.docgen import get_docgen_agent, supported_exts as docgen_supported_exts
from llmsdk.services.datagpt import get_datagpt_agent, supported_exts as datagpt_supported_exts
from llmsdk.services.extractor import get_contract_profilespec, contract_query

#####################################
# Helper Functions...
#####################################
def get_chroma_index_summary(indexname, persistdirectory):

    print(indexname)
    print(persistdirectory)
    db = Chroma(indexname,
                embedding_function=OpenAIEmbeddings(),
                persist_directory=persistdirectory)

    content = db.get()

    sample = {
        "source": "/home/pingali/Data/enrich/llm-agents/data/docs/qna-batch-brookfield/October Three RFP/converted/Ste Michelle Bid Specifications 06.21.2023.pdf",
        "file_path": "/home/pingali/Data/enrich/llm-agents/data/docs/qna-batch-brookfield/October Three RFP/converted/Ste Michelle Bid Specifications 06.21.2023.pdf",
        "page": 0,
        "total_pages": 7,
        "format": "PDF 1.5",
        "title": "",
        "author": "Emmy Brown",
        "subject": "",
        "keywords": "",
        "creator": "Writer",
        "producer": "LibreOffice 6.4",
        "creationDate": "D:20230719202934+05'30'",
        "modDate": "",
        "trapped": "",
        "bboxes": "[[264.3999938964844, ...,489.29425048828125]]",
        "file": "Ste Michelle Bid Specifications 06.21.2023.pdf",
        "chunk": 1,
        "id": "922e098a97dd5fcea7343d052317ee71"
        }

    summary = {}
    metadatas = content['metadatas']
    for entry in metadatas:
        filename = entry['file']
        page     = entry['page']
        chunk    = entry['chunk']
        format_  = entry['format']
        if filename not in summary:
            summary[filename] = {
                'pages': [],
                'chunks': [],
                'format': "unknown"
            }
        if page not in summary[filename]['pages']:
            summary[filename]['pages'].append(page)
        summary[filename]['chunks'].append(chunk)
        summary[filename]['format'] = format_

    return summary

@click.group()
def process():
    """
    Run the agent in batch mode with documents
    """

def show_help():
    helptext = """This script needs to run with

export DATA_ROOT="/home/pingali/Data/enrich/llm-agents/data/docs"
export AGENTNAME=default
    """
    print(helptext)


#####################################
# QnA Simple - single command
#####################################
@process.command("qna-simple")
@click.argument("dataset")
@click.argument("question")
def qna(dataset, question):
    """
    Ask a question against a qna agent
    """

    if 'DATA_ROOT' not in os.environ:
        show_help()
        return

    details = async_to_sync(get_agent_details)(namespace="qna",
                                               username="cli",
                                               subdir=dataset,
                                               exts=qna_supported_exts,
                                               get_task_specific_agent=get_qna_agent)

    result = details['agent'].query(question, mode="internal")
    print(json.dumps(result, indent=4))

#####################################
# qna-batch execution
#####################################
def get_path_source(fpath):
    if os.path.isdir(fpath):
        return "dir"
    else:
        return fpath.split('.')[-1]

@process.command("qna-batch")
@click.option("--input", "-i",
              required=True,
              multiple=True,
              help="Path to input docset/file")
@click.option("--queryspec", "-q",
              required=True,
              help="Path to queryspec")
@click.option("--output", "-o",
              required=True,
              help="Path to store output json")
@click.option("--platform", "-p",
              default="openai",
              help="LLM platform backend to use {openai, azure}")
@click.option("--agentname", "-a",
              required=True,
              help="Agent name")
@click.option("--indexname", "-x",
              required=True,
              help="Name of index")
@click.option("--force-reindexing/--no-force-reindexing",
              required=False,
              default=True,
              help="Whether the index should be built")
@click.option("--vectorstore", "-s",
              default="chroma",
              help="Flavour of Vector DB (chroma, faiss)")
@click.option("--persistdirectory", "-p",
              default="chromadb",
              help="Path to directory to persist vector DB")
def qna_batch(agentname, indexname,
              input, queryspec,
              output, platform, vectorstore,
              persistdirectory, force_reindexing):
    """
    Take a query spec and docset and extract information in batch mode
    """

    # create an agent
    agent = LLMQuerierExtractor(name=agentname, platform=platform)

    # Cleanup the name. max 60 chars, starts and ends with
    # alphanumeric char
    indexname = slugify(indexname)[:60]
    indexname = re.sub(r"^[^a-zA-Z0-9]+","", indexname)
    indexname = re.sub(r"[^a-zA-Z0-9]+$","", indexname)

    print("Index", indexname)
    print("Persist Directory", persistdirectory)
    print("Force reindexing", force_reindexing)

    try:

        stats = None
        create = True

        if not force_reindexing:
            try:
                agent.load_index(store=vectorstore,
                                 persist_directory=persistdirectory,
                                 index_name=indexname)
                stats = agent.get_index_stats()
                create = False
            except:
                #traceback.print_exc()
                pass
    except:
        # traceback.print_exc()
        stats = None
        create = True

    if ((stats is not None) and (stats['n_items'] == 0)):
        create = True

    if not create:
        summary = get_chroma_index_summary(indexname, persistdirectory)
        print(json.dumps(summary, indent=4))

    if create:
        print("Creating index")
        cnt = 0
        for fpath in input:
            # figure out the source
            source = get_path_source(fpath)

            # load the data
            data = agent.load_data(source=source, content=fpath)

            # add to index
            if cnt == 0:
                # only for the first file
                agent.create_add_index(data=data,
                                       store=vectorstore,
                                       persist_directory=persistdirectory,
                                       index_name=indexname)
            else:
                # for subsequent files
                agent.add_to_index(data=data)

            cnt += 1

    # load query spec
    queryspecfile = queryspec # dont confuse spec with filename of that spec
    if queryspecfile.endswith(".json"):
        with open(queryspecfile, "r") as fd:
            queryspec = json.load(fd)
    elif queryspecfile.endswith(".py"):
        module_name = "queryspecmod"
        spec = importlib.util.spec_from_file_location(module_name, queryspec)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        # Now get hold of query spec
        queryspec = module.get_queryspec()
    else:
        raise Exception(f"Unable to load the query spec: {queryspecfile}")

    # Limit the questions...for testing..
    # queryspec['query_set'] = queryspec['query_set'][:1]

    # process the spec
    responses = agent.process_spec(queryspec)

    # store
    with open(output, "w") as fd:
        json.dump(responses, fd, indent=2)
    print (f"Stored response json at: {output}")


#####################################
# DocGen
#####################################
@process.command("docgen-run")
@click.argument("dataset")
@click.argument("specification")
@click.option("--output", "-o",
              help="Output file")
def docgen(dataset, specification, output):
    """
    Generate document
    """

    if 'DATA_ROOT' not in os.environ:
        show_help()
        return

    # Load the specification
    spec = json.load(open(specification))

    # Get the agent
    details = async_to_sync(get_agent_details)(namespace="docgen",
                                               username="cli",
                                               subdir=dataset,
                                               exts=docgen_supported_exts,
                                               get_task_specific_agent=get_docgen_agent)

    # Generate the output
    answer   = details['agent'].generate_doc(spec)

    # => Now output the result..
    text = answer # answer not a json
    if output is not None:
        output = os.path.abspath(output)
        with open(output, 'w') as fd:
            fd.write(text)
        print(f"Output: {output}")
    else:
        print(text)

@process.command("docgen-sample-spec")
@click.option("--output", "-o",
              help="Output file")
def docgen_spec(output):
    """
    Generate a sample specification
    """
    profilespec = {
	    "prompt": {
		    "prefix": "write a para about",
		    "suffix": "for this startup company"
	    },
	    "sections": {
		    "about": {
			    "enable": True,
		        "title": "About the Company",
			    "intent": "introducing the startup",
			    "questions": [
				    "industry and specific focus",
				    "key technologies developed",
				    "unique value proposition"
			    ],
			    "collate": False,
		    },
		    "founders": {
			    "enable": True,
			    "title": "Company Founders",
			    "intent": "about the co-founders of the startup",
			    "questions": [
				    "co-founders and the roles they hold",
				    "past experience of co-founders"
			    ],
			    "collate": "rewrite"
		    },
		    "market-analysis": {
			    "enable": True,
			    "title": "Market Analysis",
			    "intent": "about market analysis for this startup",
		        "questions": [
			        "target customer base",
			        "other potential customers",
				    "other comparable companies and startups in this space",
				    "the main competitors",
				    "biggest risk factors",
				    "key differentiators",
				    "market trends for the domain"
			    ],
			    "collate": False
		    }
	    }
    }

    text = json.dumps(profilespec, indent=4)
    if output is not None:
        output = os.path.abspath(output)
        with open(output, 'w') as fd:
            fd.write(text)
        print(f"Output: {output}")
    else:
        print(text)

#####################################
# DataGPT
#####################################
@process.command("datagpt")
@click.argument("dataset")
@click.argument("question")
@click.option("--output", "-o",
              default=None,
              help="Output file")
def datagpt(dataset, question, output):
    """
    Interact with a dataset
    """

    if 'DATA_ROOT' not in os.environ:
        show_help()
        return

    details = async_to_sync(get_agent_details)(namespace="datagpt",
                                               username="cli",
                                               subdir=dataset,
                                               exts=datagpt_supported_exts,
                                               get_task_specific_agent=get_datagpt_agent)

    result = details['agent'].query(question)

    text = json.dumps(result, indent=4)
    if output is not None:
        output = os.path.abspath(output)
        with open(output, 'w') as fd:
            fd.write(text)
        print(f"Output: {output}")
    else:
        print(text)


#####################################
# Contract processing
#####################################
@process.command("contract-sample-spec")
@click.argument("specfile")
@click.option("--nature", "-n",
              default="contract",
              type=click.Choice(['contract'],
                                case_sensitive=False))
def contract_sample_spec(nature, specfile):
    """
    Get a sample spec for extractor

    Store as llm-runner extractor-sample-spec -n contract 2> .../work/spec.py
    """

    if 'DATA_ROOT' not in os.environ:
        show_help()
        return

    if os.path.exists(specfile):
        print("Specfile exists. Please remove first")
        print(specfile)
        return

    spec = inspect.getsource(get_contract_profilespec)
    spec += "\nget_spec = get_contract_profilespec\n"

    with open(specfile, 'w') as fd:
        fd.write(spec)

    print("Use this with extractor-run")
    print(f"llm-runner extractor-run {specfile} <source-document>")

@process.command("contract-run")
@click.argument("spec")
@click.argument("filename")
@click.option("--output", "-o",
              default=None,
              help="Output file")
def contract_extractor_run(spec, filename, output):
    """
    Run a extractor spec against the input files

    spec is generated using extractor-sample-spec
    """

    if ((os.path.abspath(filename) != filename) or
        (os.path.abspath(spec) != spec) or
        ((output is not None) and
         (os.path.abspath(output) != output))):
        print("This command required absolute paths for spec/inputfile/output")
        return

    if not os.path.exists(filename):
        print(f"Input file is missing: {filename}")
        return

    if not os.path.exists(spec):
        print(f"Profile spec file is missing: {spec}")
        return

    # Load the specification
    modspec = importlib.util.spec_from_file_location("samplemod", spec)
    mod = importlib.util.module_from_spec(modspec)
    modspec.loader.exec_module(mod)
    profilespec = mod.get_spec()

    result, metadata = contract_query(profilespec, filename)

    # output extraction
    text = json.dumps({
        "result": result,
        "metadata": metadata
    }, indent=4)

    if output is not None:
        output = os.path.abspath(output)
        with open(output, 'w') as fd:
            fd.write(text)
        print(f"Output: {output}")
    else:
        print(text)


@process.command("check-chroma-index")
@click.option("--persistdirectory", "-p",
              default="chromadb",
              help="Path to directory to persist vector DB")
@click.option("--indexname", "-x",
              required=True,
              help="Name of index")
@click.option("--output", "-o",
              required=False,
              help="Output File")
def check_chromadb(persistdirectory, indexname, output):
    """
    Check vector db
    """
    indexname = slugify(indexname)
    summary = get_chroma_index_summary(indexname, persistdirectory)

    if output is None:
        print(json.dumps(summary, indent=4))
    else:
        with open(output, 'w') as fd:
            fd.write(json.dumps(summary, indent=4))


def main():
    process()

if __name__ == "__main__":
    main()



import os
import sys
import json
import random
import copy
import hashlib
import pickle
import logging
import traceback
from collections import defaultdict

import seaborn as sns
from faker import Faker
import pandas as pd
import numpy as np

from enrichsdk.utils import SafeEncoder, note
from enrichsdk.lib import get_credentials_by_name
from enrichsdk.contrib.lib.assets import LLMTextGenerator

__all__ = [
    'BaseAnonymizer',
    'CachingAnonymizer',
    'AnonymizerMixin'
]

logger = logging.getLogger('app')

class BaseAnonymizer(object):

    def __init__(self, textgen_cred, *args, **kwargs):
        """
        default
        """
        # set vars
        self.textgen_cred = textgen_cred

        # what are our possible column classes
        self.column_classes = {
            'person': 'person-name',
            'country': 'country_2char',
            'address': 'location-address',
            'phone': 'phone_number',
            'mobile': 'phone_number',
            'email': 'email',
            'identifier': 'categorical',
            'number': 'numeric',
            'latlong': 'numeric',   # need to write a custom anonymizer
            'datetime': 'numeric',  # need to write a custom anonymizer
            'location': 'categorical',
        }

        # init a Faker object
        self.fakeObj  = self.init_faker_object()

        super().__init__(*args, **kwargs)


    def init_faker_object(self):
        fakeObj = Faker()
        names           = defaultdict(fakeObj.name)
        addresses       = defaultdict(fakeObj.address)
        countries       = defaultdict(fakeObj.bank_country)
        phonenumbers    = defaultdict(fakeObj.msisdn)
        emaildomains    = defaultdict(fakeObj.domain_name)

        obj = {
            "faker": fakeObj,
            "generators": {
                "person-name": names,
                "country_2char": countries,
                "location-address": addresses,
                "phone_number": phonenumbers,
                "email_domain": emaildomains,
            }
        }

        return obj

    ###########################################
    # Column Classification
    ###########################################
    def classify_columns_naive(self, df, columns):
        for col_name, col_obj in columns.items():

            if pd.api.types.is_numeric_dtype(df[col_name]):
                anon_type = "numeric"
            elif pd.api.types.is_string_dtype(df[col_name]):
                anon_type = "categorical"
            elif pd.api.types.is_categorical_dtype(df[col_name]):
                anon_type = "categorical"
            else:
                anon_type = "categorical"

            columns[col_name]['anon_type'] = anon_type
            columns[col_name]['values'] = df[col_name].head().values.tolist()
            columns[col_name]['dtype'] =  df[col_name].dtypes.name

        return columns

    def classify_columns_common(self, df, columns, method="scribble_llm", spec={}):

        cache = spec.get('cache',{})

        # do the column classification
        if method == "scribble_llm":
            # init the credentials

            # first, do a naive classification so we can find the non-numeric columns
            cols = self.classify_columns_naive(df, columns)
            cols_to_classify = {}
            for c, o in list(cols.items()):
                if o['anon_type'] != 'numeric':
                    if c in cache:
                        cols[c]['anon_type'] = cache[c]
                    else:
                        cols_to_classify[c] = o

            # figure out what are the columns we need to classify
            class_str   = ",".join([c for c, o in self.column_classes.items()])
            class_len   = len(self.column_classes)

            if len(cols_to_classify) > 0:
                cols_str    = ",".join([c for c, o in cols_to_classify.items()])

                prompt = f"""Assume you have {class_len} classes:
{class_str}
How would you classify the following columns in a database table:
{cols_str}"""

                cred = get_credentials_by_name(self.textgen_cred)
                generator = LLMTextGenerator(cred)
                result = generator.generate_text(prompt=prompt)
                if result['success'] == True:
                    # get the class asignments
                    classes = result['text'].split('\n')
                    for c in classes:
                        if ":" not in c:
                            continue
                        [col_name, label] = c.split(":")
                        col_name = col_name.strip().lower()
                        label = label.strip().lower()
                        cache[col_name] = columns[col_name]['anon_type'] = self.column_classes[label]

            else:
                # something went wrong, so we have to fall back on the default classification method
                columns = self.classify_columns_naive(df, columns)

        else:
            # fall back on the default classification method
            columns = self.classify_columns_naive(df, columns)

        return columns


    ###########################################
    # Data Anonymization
    ###########################################
    def anonymize_dataset(self, df, spec={}):
        """
        Anonymize a dataset given a spec.
        The spec defines how the dataset should be handled
        and what kinds of anonymization needs to be performed.
        If no spec is given, we infer one from the dataset.
        """

        name    = spec.get('name', 'dataframe-anonymizer')
        config  = spec.get('config', {})
        columns = config.get('columns', {})
        sample  = config.get('sample', -1)

        # if no columns are specified
        # we need to anonymize all available columns
        if len(columns) == 0:
            _columns = {}
            for c in df.columns:
                _columns[c] = {"include": "yes"}
            columns = _columns

        # Sanity check...
        for c, details in columns.items():
            if not isinstance(details, dict):
                raise Exception(f"column specification {c} should be a dictionary")

        # Sample if required
        if ((sample > 0) and (sample < len(df))):
            df = df.sample(sample)

        # we need to classify the columns
        # so we know what kind of anonymization to use
        columns = self.classify_columns_common(df, columns, spec)

        # prepare for anonymization
        anon_cols       = []
        missing_cols    = []
        error_cols      = []
        anon_actions    = {}
        dropped_cols    = []
        retained_cols   = []
        anon_df         = df.copy()

        # run through each column and try to anonymize
        for col_name, col_obj in columns.items():
            include = col_obj.get("include", "yes")
            if include == "yes":
                if col_name not in df.columns:
                    # this column in the spec is missing in the DF, make note of it
                    missing_cols.append(col_name)
                else:
                    try:
                        # anonymize this single column
                        anon_df[col_name], msg = self.anonymize_single_column(col_name, col_obj, df)
                        # make note of the column name
                        anon_cols.append(col_name)
                        # make note of the anonymization action
                        anon_actions[col_name] = msg
                    except Exception as e:
                        traceback.print_exc()
                        error_cols.append(col_name)

        # drop the other columns if required by spec
        nontransformed = config.get("nontransformed", "retain")
        if nontransformed == "drop":
            dropped_cols = [c for c in df.columns if c not in anon_cols]
            anon_df = anon_df[anon_cols]
        else:
            retained_cols = [c for c in df.columns if c not in anon_cols]

        # construct the result object
        anon_data = {
            "df": anon_df,
            "actions": anon_actions,
            "anon_cols": anon_cols,
            "missing_cols": missing_cols,
            "error_cols": error_cols,
            'dropped_cols': dropped_cols,
            'retained_cols': retained_cols,
        }

        success = True if len(anon_cols)>0 else False

        return success, anon_data

    def anonymize_single_column(self, col_name, col_obj, df, params={}):
        """
        Takes a dataset and anonymizes the specified column
        """

        # get the faker object
        fakeObj = self.fakeObj

        # setup handlers for the various anonymization types
        generators = {}
        # first for the lookup generators
        for g, lookup in fakeObj['generators'].items():
            generators[g] = {
                "type": "lookup",
                "handler": lookup
            }
        # then for the custom generators
        generators["numeric"] = {
            "type": "custom",
            "handler": "anon_numeric"
        }
        generators["categorical"] = {
            "type": "custom",
            "handler": "anon_categorical"
        }
        generators["email"] = {
            "type": "custom",
            "handler": "anon_email"
        }

        anon_type = col_obj.get('anon_type')

        _d = []
        if anon_type in generators:
            gen_type = generators[anon_type]['type']
            gen_handler = generators[anon_type]['handler']
            if gen_type == "lookup":
                # we call the apply only on the specific column
                data = df[col_name].apply(lambda x: gen_handler[x])
                l_msg = ""
            else:
                handler = getattr(self, gen_handler)
                # we call the apply to the full dataframe, we may need other columns
                # return is only the relevant column
                data, l_msg = handler(df, col_name, col_obj)
            msg = f"Anonymized using <{gen_handler}> handler of <{gen_type}> type." + l_msg
        else:
            data = np.nan
            msg = f"No <{anon_type}> generator found, defaulting to NaN."

        return data, msg


    #######################################
    # Custom anonymizers go here
    #######################################
    def anon_numeric(self, df, col_name, column):
        """
        Method to fuzz numeric data. Various fuzzing methods can be defined here.
        Input is the full dataframe, output is the relavant column being fuzzed.
        """
        msg = ""

        method      = column.get("method", "perturb")
        params      = column.get("params", {})

        val = df[col_name].fillna(0)

        if method == "perturb":
            range = params.get("range", 0.05)
            val += random.uniform(-range*val, range*val)
        else:
            msg = f"Unknown method to anon NUMERIC column, setting default NaNs."
            val = np.nan

        return val, msg

    def anon_categorical(self, df, col_name, column):
        """
        Method to anonymize categorical data. Various anonymization methods can be defined here.
        Input is the full dataframe, output is the relavant column being anonymized.
        """
        msg = ""

        def generate_hash(txt):
            if pd.isnull(txt):
                return np.nan

            if not isinstance(txt, str):
                txt = str(txt)

            hashed = hashlib.md5(txt.encode('utf-8')).hexdigest()
            hashed = hashed[:-10]
            return hashed

        method      = column.get("method", "hash")
        params      = column.get("params", {})

        if method == "hash":
            val = df.apply(lambda x: generate_hash(x[col_name]), axis=1)
        else:
            msg = f"Unknown method to anon CATEGORICAL column, setting default NaNs."
            val = np.nan

        return val, msg

    def anon_email(self, df, col_name, column):
        """
        Method to anonymize email data. Can generate emails to match or not match
        data in some name field. Also respects original email domain distribution if required.
        Input is the full dataframe, output is the relavant column being anonymized.
        """
        msg = ""

        match_names = column.get("match_names", True)
        if match_names is True:
            if "name_field" not in column:
                msg = f"Unknown name field to match emails, setting random emails."
                match_names = False
                return np.nan, msg
            else:
                if column["name_field"] not in data.columns:
                    msg = f"name field not in dataframe, setting random emails."
                    match_names = False
                    return np.nan, msg

        def generate_email(fakeObj, row, col_name, column, match_names):

            email     = row[email_col_name]
            if pd.isnull(email):
                return np.nan

            # whitelist of email domains
            # if the origninal email is in this list, don't replace it
            # useful to maintain data distribution
            domain_whitelist = ['gmail.com',
                                'yahoo.com',
                                'hotmail.com',
                                'aol.com']

            email_col_name  = col_name
            orig_domain     = row[email_col_name].split('@')[1]

            # set the email domain first
            if column.get("dist", "yes") == "yes":
                # we need to ensure that the distribution of generated email domains
                # match what was present in the input
                # popular free email domains will carry over, while others will be
                # replaced with random domains while still retaining distribution
                if any([d==orig_domain for d in domain_whitelist]):
                    # retain the origninal domain name
                    domain = orig_domain
                else:
                    # get a new domain name
                    domain = fakeObj['generators']['email_domain'][orig_domain]
            else:
                # no need to match distribution of generated email domains
                domain = fakeObj['faker'].ascii_email().split('@')[1]

            if match_names is True:
                # we want to match the anon email with the name field
                name = row[column['name_field']]
                names = unidecode.unidecode(name).lower().split(' ')
            else:
                # we don't care about matching the anon email with the name field
                names = fakeObj['faker'].name().split(' ')

            firstname = names[0]
            lastname = names[-1]

            # possible variations of email
            nameparts = {
                1: f"{firstname}",
                2: f"{lastname}",
                3: f"{firstname}.{lastname}",
                4: f"{firstname}.{firstname[0]}.{lastname}",
                5: f"{firstname}.{lastname[0]}.{lastname}",
                6: f"{firstname}.{firstname[0]}.{lastname[0]}",
                7: f"{firstname}.{random.randint(1,10000)}",
                8: f"{firstname}_{random.randint(1,10000)}",
                9: f"{firstname}.{lastname}.{random.randint(1,10000)}",
            }
            choice = random.randint(1, len(nameparts))
            namepart = nameparts[choice]
            email = f"{namepart}@{domain}"

            return email

        val = df.apply(lambda x: generate_email(self.fakeObj, x, col_name, column, match_names), axis=1)

        return val, msg


class CachingAnonymizer(BaseAnonymizer):
    """
    Cache results of the classification
    """

    def __init__(self, cachepath, *args, **kwargs):
        self.cachepath = cachepath
        if os.path.exists(cachepath):
            try:
                self.cache = pickle.load(open(cachepath, 'rb'))
            except:
                self.cache = {}
                logger.exception("Unable to load anonymizer cache",
                                 extra={
                                     'data': cachepath
                                 })

        else:
            self.cache = {}

        super().__init__(*args, **kwargs)

    def update_cache(self):
        with open(self.cachepath, 'wb') as fd:
            pickle.dump(self.cache,
                        fd,
                        protocol=pickle.HIGHEST_PROTOCOL)


    def anonymize_dataset(self, df, spec={}):
        """
        Filter out column that neednt be computed
        """

        spec['cache'] = self.cache

        # Run the
        success, anon_data = super().anonymize_dataset(df, spec)

        self.update_cache()

        return success, anon_data

class AnonymizerMixin(object):
    """
    Embed the core anonymization functions in transforms
    """

    def anonymize_init(self, anonargs):
        """
        Initialize the anonymizer
        """

        if not hasattr(self, 'state'):
            raise Exception("Missing state attribute in transform")

        #=> Initialize the anonymization args
        if (('cache' not in anonargs) or
            ('textgen_cred' not in anonargs)):
            logger.error("Missing cache or textgencred",
                         extra={
                             'transform': self.name,
                             'data': json.dumps(anonargs, indent=4)
                         })
            raise Exception("Missing cache or textgencred")

        self.anonymizer = CachingAnonymizer(cachepath=anonargs['cache'],
                                            textgen_cred=anonargs['textgen_cred'])


        if 'targets' not in anonargs:
            logger.error("Missing targets for anonymization",
                         extra={
                             'transform': self.name,
                             'data': json.dumps(anonargs, indent=4)
                         })
            raise Exception("Missing anonymization targets")

        # => Clean the targets and store them after including the spec
        defaultspec = anonargs.get('spec', {
            'config': {
                'sample': 1000,
                'nontransformed': 'drop'
            }
        })

        targets = anonargs['targets']
        cleaned_targets = []
        for t in targets:
            if isinstance(t, str):
                t = {
                    'name': t,
                }
            if 'name' not in t:
                logger.warning("Anonymization target skipped",
                               extra={
                                   'transform': self.name,
                                   'data': json.dumps(t)
                               })
                continue

            if 'spec' not in t:
                t['spec'] = copy.copy(defaultspec)

            cleaned_targets.append(t)

        anonargs['targets'] = cleaned_targets

        logger.debug("Anonymization targets",
                           extra={
                               'transform': self.name,
                               'data': json.dumps(anonargs)
                           })

        self.anonargs = anonargs
        self.anonymization_actions = []

    def anonymize_target(self, name, df=None):

        anon_name = f"{name}_anonymized"

        target = None
        for spec in self.anonargs['targets']:
            if spec.get("name", "unknown") == name:
                target = spec
                break

        if target is None:
            return

        # => If df is not specified, then lookup the state
        if df is None:

            if not self.state.has_frame(name):
                logger.error(f"Missing {name} in state",
                             extra={
                                 'transform': self.name,
                             })
                return

            # => Now get the state
            detail = self.state.get_frame(name)

            # => Extract the dataframe to be anonymized
            df = detail['df']

        if ((df is None) or (len(df) == 0)):
            raise Exception(f"Missing or empty {name}")

        # Anonymize
        success, anon_data = self.anonymizer.anonymize_dataset(df, spec)

        # => Dont update it again if it is already present
        if self.state.has_frame(anon_name):
            return anon_data['df']


        msg = ""

        for attr, detail in anon_data.items():
            if attr == "df":
                msg += note(anon_data['df'], f"Anonymized {name}")
            elif "cols" in attr:
                msg += f"{attr}: {','.join(detail)}\n"
            elif attr == "actions":
                for k,v in detail.items():
                    msg += f"Action: {k} => {v}\n"
        logger.debug(f"Anonymized {name}",
                     extra={
                         'transform': self.name,
                         'data': msg
                     })

        # Track the anonymization status
        for column, action in anon_data['actions'].items():
            self.anonymization_actions.append({
                'frame': name,
                'column': column,
                'action': 'anonymized',
                'detail': action
            })

        for attr, label in {
                "missing_cols": "missing",
                "error_cols": "error",
                "dropped_cols": "dropped",
                "retained_cols": "retained"
            }.items():

            cols = anon_data.get(attr, [])
            for col in cols:
                self.anonymization_actions.append({
                    'frame': name,
                    'column': column,
                    'action': label,
                    'detail': ""
                })

        #=> Register the anonymized dataframe with state...
        anon_df = anon_data['df']

        # => Gather the update parameters
        updated_detail = {
            'df': anon_data['df'],
            'description': f"{name} anonymized",
            'transform': self.name,
            'frametype': 'pandas',
            'params': self.get_column_params(anon_name, anon_df) + [
                {
                    "type": "lineage",
                    "dependencies": [
                        {
                            "type": "dataframe",
                            "nature": "input",
                            "objects": [name]
                        }
                    ]
                }
            ],
        }

        # Update the state.
        # Do the same thing for the second update dataframe
        self.state.update_frame(anon_name, updated_detail, create=True)

        # => Return the dataframe..
        return anon_data['df']

    def anonymize_all_targets(self):

        targets = self.anonargs['targets']

        if len(targets) == 0:
            logger.debug("No anonymization targets found",
                         extra=self.config.get_extra({
                             'transform': self.name
                         }))
            return

        for t in targets:
            self.anonymize_target(t['name'])

    def anonymize_finalize(self):

        allactions = self.anonymization_actions
        if len(allactions) == 0:
            logger.warning("No valid anonymization actions",
                           extra={
                               'transform': self.name
                           })
            return

        # => Gather the changes made
        action_name = "anonymization_status"
        actiondf = pd.DataFrame(allactions)
        frames = list(actiondf['frame'].unique())
        updated_detail = {
            'df': actiondf,
            'description': f"Anonymization status",
            'transform': self.name,
            'frametype': 'pandas',
            'params': self.get_column_params(action_name, actiondf) + [
                {
                    "type": "lineage",
                    "dependencies": [
                        {
                            "type": "dataframe",
                            "nature": "input",
                            "objects": frames
                        }
                    ]
                }
            ],
        }

        # Update the state.
        # Do the same thing for the second update dataframe
        self.state.update_frame(action_name, updated_detail, create=True)


if __name__ == "__main__":

    diamonddf = sns.load_dataset("diamonds")

    anonymizer  = BaseAnonymizer(textgen_cred="openai-api")
    success, anon_data  = anonymizer.anonymize_dataset(diamonddf)
    print (f"Success: {success}")
    print (json.dumps(anon_data, indent=4, cls=SafeEncoder))

    anonymizer  = CachingAnonymizer("anon.pickle",
                                  textgen_cred="openai-api")
    success, anon_data  = anonymizer.anonymize_dataset(diamonddf)
    print (f"Success: {success}")
    print (json.dumps(anon_data, indent=4, cls=SafeEncoder))




import os
import sys
import time
import json
import copy
import traceback
import logging
from urllib.parse import urlencode

import httpx
import asyncio
from asgiref.sync import sync_to_async, async_to_sync

from enrichsdk.lib import get_credentials_by_name

logger = logging.getLogger('app')

class HedwigClient():

    def __init__(self, cred):

        if isinstance(cred, str):
            self.cred = get_credentials_by_name(cred)
        else:
            self.cred = cred

    async def access_singlestep(self,
                          request_action,
                          status_action,
                          params,
                          timeout=10):
        """
        Single blocking call to the LLM
        """

        url = self.cred['url']
        q = f"{request_action}?{urlencode(params)}"
        tries = 3
        while tries > 0:
            try:
                response = httpx.get(url + q)
                break
            except httpx.TimeoutException:
                traceback.print_exc()
                logger.exception("Timeout while accessing LLM Agent")
                await asyncio.sleep(10)
            except:
                logger.exception("Error while accessing LLM Agent")

            tries -= 1
        if tries == 0:
            response = {
                'status': 'failure',
                'message': "Unable to reach LLM to start request after retries",
                '_request': params
            }
            return response

        response = response.json()
        if 'status' not in response:
            response = {
                'status': 'failure',
                'message': "Invalid response from LLM Agent: Missing status",
                '_request': params
            }
            return response

        status = response['status']
        if status != 'pending':
            status['_request'] = params
            return response

        if 'request_id'not in response:
            response = {
                'status': 'failure',
                'message': "Invalid response from LLM Agent: Missing request id",
                '_request': params
            }
            return response

        request_id = response['request_id']
        if not isinstance(request_id, str) or len(request_id) == 0:
            response = {
                'status': 'failure',
                'message': "Invalid response from LLM Agent: Invalid request id",
                '_request': params
            }
            return response

        tries = 20
        client = httpx.AsyncClient()
        try:
            while tries > 0:

                await asyncio.sleep(5)

                q = f"{status_action}?request_id={request_id}"
                # print("Calling async client get", url + q)
                try:
                    response = await client.get(url + q)
                except httpx.ReadTimeout:
                    # print("Read timeout")
                    continue
                except httpx.TimeoutException:
                    # Happens first time with cold agent
                    # print("Timeout Exception")
                    if tries == 20: # first time
                        continue
                    else:
                        return {
                            "request_id": None,
                            "status": "failure",
                            "message": "Unable to reach after retries. Agent could be busy",
                            "_params": params
                        }

                response = response.json()
                # print(f"Try {tries}", response)
                status = response.get('status', 'failure')

                # print("Found status", status)
                # print("Returning")
                # print(json.dumps(response, indent=4))
                if status != 'pending':
                    response = response['data']
                    response['_request'] = params
                    return response

                tries -= 1
                # print("Sleeping for 5 seconds")
                await asyncio.sleep(timeout)

        except httpx.TimeoutError:
            logger.exception("Timeout while accessing LLM")
            traceback.print_exc()
        except:
            logger.exception("Error while accessing LLM")
            traceback.print_exc()
        finally:
            await client.aclose()

        # print("Returning response")
        return {
            "request_id": None,
            "status": "failure",
            "answer": "Internal error while accessing LLM Agent. See log",
            '_request': params
        }

    async def access_multistep_status(self, action, params):

        url = self.cred['url']
        q = f"{action}?{urlencode(params)}"
        request_id = params.get('request_id', None)
        try:
            if request_id is None:
                return {
                    "request_id": None,
                    "status": "failure",
                    "query": "",
                    "answer": "Invalid request to LLM Agent. Missing request_id",
                    "_request": params
                }

            print("Calling async client get", url + q)

            try:
                client = httpx.AsyncClient()
                response = await client.get(url + q)
                response = response.json()
                data = response.pop('data', {})
                response.update(data)
                response['request_id'] = request_id
                if 'answer' not in response:
                    response['answer'] = "No answer received"
                if 'query' not in response:
                    if 'question' in response:
                        response['query'] = response['question']
                    else:
                        response['query'] = ''

                return response
            except:
                logger.exception("Error while accessing LLM")
                traceback.print_exc()
                return {
                    "request_id": request_id,
                    "status": "pending",
                    "query": "",
                    "answer": "Unable to reach LLM Agent",
                    "_request": params
                }
            finally:
                await client.aclose()
        except:
            logger.exception("Unable to reach LLM Agent. See log")
            return {
                "request_id": None,
                "status": "failure",
                "query": "",
                "answer": "Internal error while accessing LLM Agent. See log",
                "_request": params
            }

    async def access_multistep_initiate(self, action, params):


        url = self.cred['url']
        q = f"{action}?{urlencode(params)}"

        # => Handle the cold-start when the document is loading and the agent is blocked...
        tries = 3
        while tries > 0:
            try:
                response = httpx.get(url + q)
                break
            except httpx.TimeoutException:
                await asyncio.sleep(10)
                tries -= 1

        if tries == 0:
            response = copy.copy(params)
            response.update({
                'response_id': None,
                'status': "failure",
                "answer": "Unable to reach LLM to start request after retries"
            })
        else:
            response = response.json()

        # Now we got a response...
        # print(json.dumps(response, indent=4))
        if ((response is None) or
            (not isinstance(response, dict)) or
            (len(response) == 0)):
            logger.error("Invalid response from agent",
                         extra={
                             'data': str(response)
                         })
            response = {
                'response_id': None,
                'status': "failure",
                "query": query,
                "answer": "Invalid response from LLM Agent"
            }

        status = response.get('status', None)
        answer = response.get('answer', None)
        if ((answer is None) and (status != 'pending')):
            logger.error("Received no answer",
                         extra={
                             'data': json.dumps(response, indent=4)
                         })
            answer = response['answer'] = "No answer. Possibly an internal error"

        return response


def test_hedwig_singlestep():
    client = HedwigClient('llm-default')

    response = async_to_sync(client.access_singlestep)(request_action="/qna",
                                                       status_action="/qna/status",
                                                       params={
                                                           "user": "venkata",
                                                           "dataset": "verizon",
                                                           "mode": "internal",
                                                           "query": "What is this document about?",
                                                           "namespace": "default",
                                                           "context": ""


                                                       })
    print(json.dumps(response, indent=4))

def test_hedwig_multistep():
    client = HedwigClient('llm-default')

    response = async_to_sync(client.access_multistep_initiate)("/qna",
                                                               params={
                                                                   "user": "venkata",
                                                                   "dataset": "verizon",
                                                                   "mode": "internal",
                                                                   "query": "What is this document about?",
                                                                   "namespace": "default",
                                                                   "context": ""
                                                       })
    print(json.dumps(response, indent=4))

    request_id = response['request_id']

    tries = 20
    while tries > 0:
        time.sleep(5)
        response = async_to_sync(client.access_multistep_status)("/qna/status",
                                                               params={
                                                                   "request_id": request_id
                                                               })
        print(json.dumps(response, indent=4))
        status = response.get('status', None)
        if status == 'pending':
            tries -= 1
            continue
        else:
            break

if __name__ == "__main__":

    #test_hedwig_singlestep()
    test_hedwig_multistep()



def has_permissions(user, orgname, role=None):
    """
    Check whether user has permissions
    """

    if not user.is_authenticated:
        return False

    if user.is_superuser:
        return True

    orgroles = list(user.org_role_users.all())

    # Filter out all other organizations
    orgroles = [o for o in orgroles if o.org.name == orgname]

    # For the specific organization
    roles = [o.name for o in orgroles]

    if ("admin" in roles) or (role is not None and role in roles):
        return True

    # Denied
    return False



#!/usr/bin/env python3
"""

Script to extract tables from pdfs or image files. This will be piped through the textract 

"""
import os
import sys
import json
import tempfile
import click

from llmsdk.lib.extractor import extract_from_file

@click.group()
def process():
    """
    Extract images/text from documents
    """

@process.command()
@click.argument("filename")
@click.argument("output")
@click.option("--provider",
              default="aws",
              help="Extractor implementation to use")
def extract(filename, output, provider):
    """
    Extract tables from filename

    PDF and imag formats supported
    """

    allresults = extract_from_file(filename, provider)
    
    with open(output, 'w') as fd:
        fd.write(json.dumps(allresults, indent=4))
    print(f"Output is in {output}")

def main():
    process()

if __name__ == "__main__":
    main()




from fastapi.concurrency import run_in_threadpool

from .basegpt import BaseGPT
from .lib import *
from ..agents import LLMQueryMapper

logger = get_logger()


class QueryMapperGPT(BaseGPT):

    def __init__(self):
        super().__init__()
        self.name = "QueryMapperGPT"
        self.agentname = "querymappergpt"
        self.platform = "openai"
        self.namespace = "querymappergpt"

    async def get_llm_agent(self):
        querymapper_agent = LLMQueryMapper(name=self.agentname, platform=self.platform)

        logger.debug(f"Built agent for {self.agentname} and platform {self.platform}")

        return {
            "agent": querymapper_agent
        }

    def startup_extra_steps(self):
        spec = self.get_customer_profile_spec()

        self.agent.load_spec(spec=spec,
                             persist_directory=self.persist_directory,
                             index_name=self.index_name)

    def get_customer_profile_spec(self):
        # ruleset
        RULE__TIME_PERIOD = "TIME_PERIOD must be one of 'day', 'week', 'month', 'quarter', or 'year'. Previous three months should be interpreted as a quarter."
        RULE__COUNTRY = "All COUNTRY names must be converted to 2 character ISO country codes"
        RULE__METRIC = "METRIC must be one of 'NSV' or 'NR'"

        # setup the template spec
        default = {
            "name": "acme_gpt",
            "topk": 5,
            "context": [
                "these queries are about cross-border financial transactions (traffic) at a fintech company",
                "'financial institutions' are also known as banks",
                "GSV is a metric measuring sales volume",
                "NSV is a metric measuring sales volume",
                "NR is a metric measuring revenue"
            ],
            "templates": [
                {
                    "template": "how many partners went live over the past TIME_PERIOD",
                    "metadata": {
                        "key": "new_partners|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what new B2B partners went live over the past TIME_PERIOD",
                    "metadata": {
                        "key": "new_b2b_partners|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "which new source and/or destination countries have been launched this TIME_PERIOD",
                    "metadata": {
                        "key": "new_countries|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what is the trend of SOURCE_COUNTRY to DESTINATION_COUNTRY traffic over the last TIME_PERIOD",
                    "metadata": {
                        "key": "traffic_trend|SOURCE_COUNTRY|DESTINATION_COUNTRY|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD,
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "what is the list of all partners in COUNTRY",
                    "metadata": {
                        "key": "partners|COUNTRY",
                        "rules": [
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "what is the list of all source partners in COUNTRY",
                    "metadata": {
                        "key": "source_partners|COUNTRY",
                        "rules": [
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "what is the list of all destination partners in COUNTRY",
                    "metadata": {
                        "key": "dest_partners|COUNTRY",
                        "rules": [
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "what is the list of 'financial institutions' in COUNTRY",
                    "metadata": {
                        "key": "banks|COUNTRY",
                        "rules": [
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "how many senders and/or receivers have been serviced this TIME_PERIOD",
                    "metadata": {
                        "key": "active_sndrs_rcvrs|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what is the expected/projected sales metrics for this TIME_PERIOD",
                    "metadata": {
                        "key": "expected_sales|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what are the total sales for this TIME_PERIOD",
                    "metadata": {
                        "key": "current_sales|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what is the total revenue for this TIME_PERIOD",
                    "metadata": {
                        "key": "current_sales|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "rank destination countries based on METRIC from SOURCE_COUNTRY",
                    "metadata": {
                        "key": "rank_dest_countries_from_source_country|METRIC|SOURCE_COUNTRY",
                        "rules": [
                            RULE__METRIC,
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "rank source partners based on METRIC from SOURCE_COUNTRY",
                    "metadata": {
                        "key": "rank_source_partners_from_source_country|METRIC|SOURCE_COUNTRY",
                        "rules": [
                            RULE__METRIC,
                            RULE__COUNTRY
                        ],
                    }
                },
                {
                    "template": "what was the sales volumes generated from new partners this TIME_PERIOD",
                    "metadata": {
                        "key": "sales_new_partners|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "from how many countries were transactions received this TIME_PERIOD",
                    "metadata": {
                        "key": "source_countries|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "to how many countries were transactions sent this TIME_PERIOD",
                    "metadata": {
                        "key": "dest_countries|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what are the highlights of business performance this TIME_PERIOD vs. last TIME_PERIOD",
                    "metadata": {
                        "key": "highlights|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what are the destination countries which were inactive over the last TIME_PERIOD",
                    "metadata": {
                        "key": "inactive_destination_countries|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what are the destination partners which were inactive over the last TIME_PERIOD",
                    "metadata": {
                        "key": "inactive_destination_partners|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what are the source countries that were inactive over the last TIME_PERIOD",
                    "metadata": {
                        "key": "inactive_source_countries|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
                {
                    "template": "what are the source partners who were inactive over the last TIME_PERIOD",
                    "metadata": {
                        "key": "inactive_source_partners|TIME_PERIOD",
                        "rules": [
                            RULE__TIME_PERIOD
                        ],
                    }
                },
            ]
        }

        return default


app = FastAPI()

##############################################
# Common
##############################################
add_show_cache(app)
add_show_health(app)
queryMapperGpt = QueryMapperGPT()
app.include_router(queryMapperGpt.router)

# # For IDE Debugging
# if __name__ == "__main__":
#     uvicorn.run(app, host="0.0.0.0", port=10892)



"""
Utilities meant for the dashboard apps
"""
import sys
import inspect
import copy
import random
import string
from ..lib.customer import find_usecase
from django.apps import AppConfig

def get_default_configuration():
    """
    Default description for any app
    """
    return """
    This app requires low code configuration implemented using the app
    sdk or other means. Please get in touch with the platform manager.
    """

def get_default_resources():
    """
    Default resources for any app
    """
    return """
    """

class EnrichAppConfig(AppConfig):
    """
    Base class for Dashboard Apps. Each is an instance of Django App.
    """

    name="default"
    """
    Name of the app
    """

    category = "default"
    """
    Category to which the app belongs
    """

    version = "v1"
    verbose_name = "default"
    description = "default"
    filename = None
    enable = False
    multiple = False
    composition = False
    status = "active"
    entry = "index"
    include_custom_urls = False
    tags = ["store"]
    _readme = ""
    _configuration = ""
    _resources = ""

    @classmethod
    def get_readme(cls):
        return cls._readme

    @classmethod
    def get_resources(cls):
        return cls._resources

    @property
    def readme(cls):
        if hasattr(cls, 'get_readme'):
            return cls.get_readme()
        elif cls._readme == "":
            return cls.description
        else:
            return cls._readme

    @property
    def configuration(cls):
        if hasattr(cls, 'get_configuration'):
            return cls.get_configuration()
        elif cls._configuration == "":
            return get_default_configuration()
        else:
            return cls._configuration

    @property
    def resources(cls):
        """
        Attribute to return the resources (a html text string)
        """
        if hasattr(cls, 'get_resources'):
            return cls.get_resources()
        elif cls._resources == "":
            return get_default_resources()
        else:
            return cls._resources

    @property
    def instanceid(self):
        """
        Unique id associated with the app
        """
        return str(id(self))

    def get_usecase(self):
        """
        Usecase to which the app belongs
        """
        if self.filename is None:
            return {}

        usecase = find_usecase(self.filename)
        self.usecase = copy.copy(usecase)
        return self.usecase

    def is_composition(self):
        return self.composition

    def get_name(self):
        return self.name

    def get_verbose_name(self):
        return self.verbose_name

    def get_description(self):
        return self.description

    def is_enabled(self):
        return self.enable

    def __str__(self):
        return f"{self.name}: {self.verbose_name}"

##################################################
# App helpers
##################################################
def generate_id(name):
    if len(name) > 0:
        return "".join([c if c.isalnum() else "_" for c in name]).lower()
    else:
        chars = string.ascii_uppercase + string.digits
        return ''.join(random.choices(chars, k=10))

def clean_and_validate_widgets(widgets):

    for w in widgets:
        if 'id' not in w:
            w['id'] = generate_id(w['name'])





"""

This module implements the QnA microservice. Its purpose is to wrap
around the QnA agent and provide an integration point from end applications.
The QnA module enables users to ask questions on a set of documents in an interactive chat-like interface.

"""
import os
import sys
import re
import glob
import yaml
import json
import traceback
import time
import hashlib
import struct
import logging
from logging.config import dictConfig
from uuid import uuid4
from datetime import datetime

from fastapi import FastAPI, BackgroundTasks, Depends
from fastapi.concurrency import run_in_threadpool
from fastapi_utils.tasks import repeat_every

from llmsdk.agents.qna import LLMIndexQuerier
from llmsdk.lib import SafeEncoder

from .log import *
from .lib import *

# Override default uvicorn logger...
logging.config.dictConfig(log_config)
logger = get_logger()

supported_exts = ['txt', 'htm', 'html', 'pdf', 'docx', 'doc']

app = FastAPI()

# Insert policy
policyspec = []
add_policy(app, policyspec)

async def get_qna_agent(namespace, username,
                        dataset, fullpath, files,
                        metadata,
                        drop_index=False):

    return await get_generic_agent(LLMIndexQuerier,
                                   dataset,
                                   fullpath,
                                   files,
                                   metadata,
                                   drop_index=False)

@app.get("/cache")
def cache_health():
    """
    Summarize the content of the cache
    """

    cache = get_cache()
    summary = defaultdict(int)
    for k, v in cache.items():
        summary[v['status']] += 1

    return summary

@app.get("/health")
def health():
    """
    Return usage statistics

    Returns
    -------
    stats: dict
           Usage statistics

    """
    logger.info("Returning stats",
                extra={
                    'data': json.dumps(stats, indent=4)
                })
    return stats


def query_update_result(request_id, result):

    # First get the params
    cache = get_cache()
    value = cache[request_id]

    for k, v in result.items():
        value[k] = result[k]

    status = result.get('status', 'unknown')
    logger.debug(f"Result: {status}",
                extra={
                    "request_id": request_id,
                    "source": "service",
                    "user": value.get("user", "unknown"),
                    "dataset": value.get("dataset", "unknown"),
                    'data': json.dumps(result, indent=4, cls=SafeEncoder)
                })

async def query_run(request_id):
    """
    Query the agent instance created

    Parameters
    ----------
    request_id: str
                UUID generated for the request

    Returns
    -------
    None
         Nothing is returned. Only the result cache is updated

    """

    cache = get_cache()

    if request_id not in cache:
        logger.error(f"Failure",
                     extra={
                         'request_id': "invalid",
                         "source": "service",
                     })
        cache[request_id] = {
            'status': "failure",
            "message": f"Invalid request id"
        }
        return

    # First get the params
    value = cache[request_id]

    try:
        params  = value['params']
        user    = params['user']
        dataset = params['dataset']
        query   = params['query']
        mode    = params['mode']
        context = params['context']
        policy  = params['policy']
        namespace = params['namespace']
        label = f"{user}_{dataset}"

        logger.debug(f"{label}: Getting agent",
                     extra={
                         "source": "service",
                         "user": user,
                         "dataset": dataset,
                         "request_id": request_id,
                         'data': json.dumps(value, indent=4, cls=SafeEncoder)
                     })

        stats['query_count'] += 1

        # First get the agent...
        details = await get_agent_details(namespace, user,
                                          dataset,
                                          exts=supported_exts,
                                          get_task_specific_agent=get_qna_agent)
        if details is None:
            query_update_result(request_id, {
                'status': "failure",
                "message": f"LLM Agent not found for {label}. See agent log",
                "result": {}
            })
            return

        agent    = details['agent']
        metadata = details['metadata']

        # Now query...
        if mode in [ "related", "similar", "docsearch"]:
            response = agent.get_similar_docs(query, topk=10)
            result = {
                "request_id": request_id,
                "question": query,
                "answer": "Relevant sections from the source are shown below:",
                "sources": []
            }

            for r in response:
                result['sources'].append({
                    "content": r.page_content,
                    "source": r.metadata['source']
                })

        else:
            agent_policy = {
                "reset_memory": policy.get('runtime',{}).get('clear_agent_memory', False)
            }
            result   = await run_in_threadpool(lambda: agent.query(query, mode=mode,policy=agent_policy))
            result['request_id'] = request_id

        logger.debug(f"Raw result received",
                     extra={
                         "source": "service",
                         "request_id": request_id,
                         "user": user,
                         "dataset": dataset,
                         "data": json.dumps(result, indent=4, cls=SafeEncoder)
                     })

        if mode in ['internal', 'economy', 'power']:
            result['answer'] = result['answer'].strip()
        elif mode == 'suggest':
            if (('suggest' not in result) or
                (not (isinstance(result['suggest'], list)))):
                answer = result['answer']
                if isinstance(answer, str):
                    suggest = re.split(r"\n|\\n", answer)
                else:
                    suggest = [s.strip() for s in answer if len(s.strip()) > 0]
                    result['suggest'] = suggest
            result['answer'] = "Here are some suggestions"
        elif ((isinstance(result['answer'], dict)) and
              ('text' in result['answer'])):
            result['answer'] = result['answer']['text']
        elif isinstance(result['answer'], dict):
            text = ""
            for k,v in result['answer'].items():
                text += f"[{k}] {v}\n"
            result['answer'] = text
        else:
            result['answer'] = str(result['answer'])

        if 'metadata' not in result:
            result['metadata'] = {}
        result['metadata'].update(metadata)
        stats['query_success'] += 1
        stats['datasets'][label]['query_count'] += 1
        stats['datasets'][label]['query_success'] += 1
        query_update_result(request_id, {
            "status": "success",
            "result": result
        })
    except Exception as e:
        stats['query_failure'] += 1
        stats['datasets'][label]['query_count'] += 1
        stats['datasets'][label]['query_failure'] += 1
        logger.exception(f"Failed to run query",
                         extra={
                             "source": "service",
                             'request_id': request_id,
                             "user": params.get('user', "unknown"),
                             "dataset": params.get('dataset', "unknown"),
                         })
        query_update_result(request_id, {
            "status": "failure",
            "message":  f"Unable to construct the answer. Internal error. See the agent log",
            "result": {
                "question": query,
                "answer": "Error while accessing the LLM Agent. See log",
                "raw": [
                    str(e)
                ]
            }
        })



@app.get("/qna/status")
async def query_status(request_id: str):
    """
    Get the status of the query request

    Parameters
    ----------
    request_id: str
                UUID generated for the request

    Returns
    -------
    status: dict
         Dictionary with request id, status, message and data

    """
    # First get the params
    try:

        value: dict = {}

        cache = get_cache()

        if request_id not in cache:
            logger.error(f"Query Status: Invalid Request ID",
                         extra={
                             'request_id': request_id,
                             "source": "service",
                         })
            return {
                "request_id": request_id,
                "status": "failure",
                "message": "Request ID not found"
            }

        value = cache[request_id]
        status = value['status']
        response = {
            "request_id": request_id,
            "status": value['status'],
            'message': value.get('message', ""),
            'data': value.get('result',{})
        }
        logger.debug(f"Query Status: {status}",
                     extra={
                         "source": "service",
                         'request_id': request_id,
                         "dataset": value.get('dataset', 'unknown'),
                         "user": value.get('user', 'unknown'),
                         'data': json.dumps(response,indent=4, cls=SafeEncoder)
                     })
        return response

    except Exception as e:
        logger.exception("Query Status: failure",
                     extra={
                         "source": "service",
                         'request_id': request_id,
                         "user": value.get('user',"unknown"),
                         "dataset": value.get('dataset',"unknown"),
                     })

        return {
            "request_id": request_id,
            "status": "failure",
            "message": str(e),
            "data": {}
        }

@app.post("/qna")
async def query_index(user: str,
                      dataset: str,
                      query: str,
                      background_tasks: BackgroundTasks,
                      mode: str = 'internal',
                      context: str = "",
                      namespace="default",
                      policy: dict = {},
                      extra: dict = {}
):
    """
    Initiate a new request

    Parameters
    ----------
    user: str
          User (an unintepreted string)
    query: str
           A string
    namespace: str
           A namespace associated with this instance of the service. Multiple
    context: str
           Any additional text
    policy: dict
           List of policies to enforce
    extra: dict
           Pass any other information

    Returns
    -------
    status: dict
         Dictionary with request id, query, status and data
    """

    cache = get_cache()
    request_id = str(uuid4())

    params = {
        "user": user,
        "dataset": dataset,
        "query": query,
        "mode": mode,
        "context": context,
        "namespace": namespace,
        "policy": policy
    }
    label = f"{user}_{dataset}"

    logger.debug(f"Starting query",
                extra={
                    "source": "service",
                    "user": user,
                    "dataset": dataset,
                    "data": json.dumps(params, indent=4, cls=SafeEncoder)
                })

    cache[request_id] = {
        "status": "pending",
        "user": user,
        "dataset": dataset,
        "params": params
    }

    # Run the background task...
    background_tasks.add_task(query_run, request_id)

    return {
        "request_id": request_id,
        "status": "pending",
        "data": {}
    }

@app.on_event("startup")
def app_startup():

    initialize_stats()



from datetime import timedelta

def static_path(match, start, end):
    """
    Time independent path. Params can specify a
    subdir

    Example::
         {
           "generate": "static_path",
           "params": { "subdir": "complete"}
         }

    """

    params = match.get('params', {})
    subdir = params.get('subdir', '.')

    return [{ 'timestamp': subdir, 'name': subdir}]

def generate_datetime_daily(match, start, end, params={}):
    """
    Given a start and an end, generate the datetime
    objects corresponding to each run for each day.

    Args:
       match (dict): has 'pattern' and generator function
       start (datetime): Start of the time range
       end (datetime): End of range. If None, default = start
    """

    params = match.get('params', {})
    pattern = match['pattern']

    if end < start:
        start, end = end, start

    names = []
    current = start
    while current <= end:
        if isinstance(pattern, str):
            name = current.strftime(pattern)
        elif callable(pattern):
            name = pattern(current)
        else:
            raise Exception("Unsupported pattern type")

        names.append({"timestamp": current.isoformat(), "name": name})
        current += timedelta(days=1)

    return names

handlers = {
    "static_path": static_path,
    "generate_datetime_daily": generate_datetime_daily,
}



import math
import ruptures as rpt
from collections import defaultdict
import matplotlib.pyplot as plt
from matplotlib.dates import date2num, num2date
from itertools import cycle
from ruptures.utils import pairwise

plt.style.use('seaborn-darkgrid')

__all__ = ['BaseChangePointDetectorModel']

class BaseChangePointDetectorModel(object):
    """
    Base class for change point detection
    """

    def __init__(self, df, *args, **kwargs):
        """
        default
        """
        # setup defaults
        self.rpt_model      = 'rbf'
        self.rpt_penalty    = 1
        self.rpt_jump       = 1
        self.rpt_minsize    = 3
        self.rpt_default_hybrid_model = 'maxvote'

        # set the dataframe
        self.data = df

        # init the change points
        self.changepoints = None

        # init the visualization
        self.viz = None

        super().__init__(*args, **kwargs)

    def detect_changepoints(self, method="hybrid"):
        """
        detect change points using the method specified
        """

        if method == "pelt":
            # use the Ruptures changepoint detection method
            self.changepoints = self.detect_changepoints_pelt()
        elif method == "hybrid":
            # use the Ruptures changepoint detection method
            self.changepoints = self.detect_changepoints_hybrid()
        else:
            # we have no other methods available
            pass

        return self.changepoints

    def get_data(self):
        """
        return the data being used
        """
        return self.data

    def get_changepoints(self):
        """
        return the detected changepoints if any
        """
        return self.changepoints


    def detect_changepoints_pelt(self,
                                    model=None,
                                    penalty=None,
                                    jump=None,
                                    min_size=None
                                    ):
        """
        run the change point detector using the Pelt method
        """
        # get the data
        ts = self.data.values

        # set the params
        if model == None:
            model = self.rpt_model
        if penalty == None:
            penalty = self.rpt_penalty
        if jump == None:
            jump = self.rpt_jump
        if min_size == None:
            min_size = self.rpt_minsize

        # fit and find change points
        algo = rpt.Pelt(model=model, jump=jump, min_size=min_size).fit(ts)
        bkps = algo.predict(pen=penalty)

        self.changepoints = bkps

        return self.changepoints

    def detect_changepoints_hybrid(self,
                                    strategy=None,
                                    model=None,
                                    penalty=None,
                                    jump=None,
                                    min_size=None
                                    ):
        """
        run the change point detector using the Hybrid method with one of the following strategies
            strict: all models should agree on the changepoints
            maxvote: a majority of models evaluated should agree on the changepoints (default strategy)
            anyone: all changepoints detected by any model are included
            select: the specified model is run
        """

        # set the params
        if strategy == None:
            strategy = self.rpt_default_hybrid_model

        ## first, run the 3 models
        changepoints = {}
        for m in ['rbf', 'l2', 'l1']:
            changepoints[m] = self.detect_changepoints_pelt(model=m)

        ## then, run the hybrid strategy
        if strategy == 'strict':
            allpoints = []
            for cp in changepoints:
                allpoints.append(changepoints[cp])
            common_bkps = sorted(list(set.intersection(*map(set,allpoints))))
        elif strategy == 'maxvote':
            allpoints = defaultdict(int)
            for model, cps in changepoints.items():
                for cp in cps:
                    allpoints[cp] += 1
            threshold = math.ceil(len(changepoints)/2)
            common_bkps = [cp for cp in allpoints if allpoints[cp]>=threshold]
        elif strategy == 'anyone':
            allpoints = []
            for model, cps in changepoints.items():
                allpoints += cps
            common_bkps = sorted(list(set(allpoints)))
        elif strategy == 'select':
            common_bkps = changepoints[model]
        else:
            # no strategy specified
            common_bkps = []

        common_bkps = sorted(common_bkps)

        self.changepoints = common_bkps

        return self.changepoints


    def visualize_changepoints(self, data, changepoints, title):
        # colors = {
        #     1: 'red',
        #     -1: 'blue',
        # }
        # COLOR_CYCLE = ["#4286f4", "#f44174"]
        #
        # ts = data.values
        # daterange = data.index
        # start_num = date2num(daterange[0].date())
        #
        # #rpt.display(ts, changepoints, figsize=(10, 3))
        # # The below code is simplified version of rpt.display()
        # # We had to modify rpt.display()to support the daterange as xaxis
        #
        # figsize = (15,7)
        #
        # fig, ax = plt.subplots(figsize=figsize)
        # ax.plot(daterange, ts)
        #
        # # color each (true) regime
        # bkps = [0] + sorted([start_num] + changepoints)
        # alpha = 0.2  # transparency of the colored background
        #
        # color_cycle = cycle(COLOR_CYCLE)
        #
        # for (start, end), col in zip(pairwise(bkps), color_cycle):
        #     ax.axvspan(max(0, start - 0.5),
        #                 end - 0.5,
        #                 facecolor=col, alpha=alpha)
        #
        # xlim = [daterange[0].date(), daterange[-1].date()]
        # ax.set_xlim(date2num(xlim))
        # ax.xaxis_date()
        #
        # fig.autofmt_xdate()
        #
        # sgn = 1
        # for p in changepoints[:-1]:
        #     style = dict(size=10, color=colors[sgn])
        #     plt.text(p+ start_num, ts[p]+(sgn),
        #              data.index[p].strftime("%-d %B, %Y"),
        #              **style)
        #     sgn *= -1
        # plt.subplots_adjust(top=.9)
        #
        # plt.title(f"Change Points: {title}")
        #
        # # save the image data
        # self.viz = plt

        colors = {
            1: 'red',
            -1: 'blue',
        }
        ts = data.values
        rpt.display(ts, changepoints, figsize=(10, 3))
        sgn = 1
        for p in changepoints[:-1]:
            style = dict(size=10, color=colors[sgn])
            plt.text(p, ts[p]+(sgn),
                     data.index[p].strftime("%-d %B, %Y"),
                     **style)
            sgn *= -1
        plt.subplots_adjust(top=.9)
        plt.title(f"Change Points: {title}")

        # save the image data
        self.viz = plt

        return self.viz



profile = {
    "sources": [
	    {
	        "nature": "db",
	        "name": "kyc_data",
	        "uri": "sqlite:///%(data_root)s/shared/aml/amldb.sqlite",
	        "query": "<query>",
	        "pipeline": [
		        "cleaner_query_result"
	        ]
	    },
        {
            "nature": "custom",
            "name": "peformance",
            "generate": "get_dataset_generic",
            "dataset": "perfdata",
            "filename": "data.csv",
            "params": {
                "low_memory": False,
                "usecols": [
                    "txn_date",
                    "status",
                    "gsv"
                ],
                "dtype": {
                    "status": "str",
                    "txn_date": "str",
                    "gsv": "float"
                }
            }
        }        
    ],
    "specs": [
	    {
            "name": "Engagement-Comparison",
	        "description": "Compare how lifetime/txn_velocity etc. compare across various dimensions",
            "sources": [
                "kyc_data",
                "perfdata",
            ],
		    "generate": "generate_result",
	        "pipeline": [
		        "store_result"
	        ]
	    }
    ]
}

def get_profile_spec():
    return profile 



"""
Generate synthetic data
"""
import json
import random

# Generating a more realistic dataset with varied company names and details

# List of sample company names
companies = [
    "AlphaTech Solutions", "Beta Enterprises",
    "Gamma Global", "Delta Dynamics", "Epsilon Electronics",
    "Zeta Industries", "Eta Energy", "Theta Transport",
    "Iota Informatics",
    "Kappa Consulting", "Lambda Labs", "Mu Manufacturing",
    "Nu Networks", "Xi Xerography", "Omicron Optics", "Pi Productions",
    "Rho Robotics", "Sigma Systems", "Tau Technologies",
    "Upsilon Utilities"
]

opportunities = []
for i, name in enumerate(companies, start=1):
    month = random.choice(range(1, 12))
    day = random.choice(range(1, 28))
    opportunity = {
        "Id": str(i),
        "Opportunity Name": f"Service Contract - {name}",
        "Account Name": name,
        "CreatedDate": f"2024-{str(month).zfill(2)}-{str(day).zfill(2)}",
        "Stage": "Negotiation" if i % 2 == 0 else "Closed Won",
        "Amount": 100000 + i * 10000,
        "Type": "New Business" if i <= 10 else "Renewal",
        "Lead Source": "Referral" if i % 3 == 0 else "Partner",
        "Next Step": "Finalize terms" if i % 4 == 0 else "Execute contract",
        "Probability (%)": 50 + i * 2,
        "Description": f"Providing {name.split()[0]} services, payment terms: {30 + i} days."
    }
    opportunities.append(opportunity)

    
def get_sample_opportunities():
    return {
        "records": opportunities
    }

def get_opportunity_descriptions():
    return { k: k for k in opportunities.keys()}




import os
import sys
import json
import yaml
import copy
import re
import traceback
import logging
from jinja2 import Template

from os.path import exists, join
import pandas as pd

# Connectors...
import pymongo
from sqlalchemy import create_engine

from ..lib.exceptions import *
from .node import *
from .mixins import *

logger = logging.getLogger("app")

__all__ = [
    "Source",
    "Sink",
    "Trigger",
    "Compute",
    "TransformSpec",
    "NotificationIntegration",
    "DatabaseIntegration",
]

##################################################
# Transform objects
##################################################
class Source(Transform):
    """
    This is a special transform that introduces dataframes into the
    state.

    This transform typically extracts information from thirdparty
    tools and services such as APIs, databases.
    """

    def __init__(self, *args, **kwargs):
        super(Source, self).__init__(*args, **kwargs)
        self.roles_supported = ["Source"]
        self.roles_current = "Source"

    def sample_inputs(self):
        return []

    def is_source(self):
        return True


class Sink(Transform):
    """
    This is a special transform that dumps dataframes that are in the
    state to some thirdparty service such as file system, database,
    cloud storage etc.
    """

    def __init__(self, *args, **kwargs):
        super(Sink, self).__init__(*args, **kwargs)
        self.roles_supported = ["Sink"]
        self.roles_current = "Sink"

    def is_sink(self):
        return True


class Compute(Transform):
    """

    This is a typical transform that wrangles data. It could introduce
    new frames and transform existing frames.

    This will mainly be pandas code but may also be spark/other
    code. The framework itself is agnostic to what the transform
    does.

    """

    def __init__(self, *args, **kwargs):
        super(Compute, self).__init__(*args, **kwargs)
        self.roles_supported = ["Compute"]
        self.roles_current = "Compute"

class TransformSpec(Transform):
    """
    Transform that supports a specification-based configuration. This
    transform requires spec.json file to be available in the transform
    directory.  The main value of this subclass is to eliminate the
    boiler plate that is required for every transform.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Load the transform specification
        self.transform_spec = self.load_tspec()

        # Load the queries...
        self.transform_input_queries = {}

        # Actions
        self.validate_tspec()
        self.load_input_queries()

        # Store the name and description
        for col in ['name', 'description']:
            setattr(self, col, self.transform_spec.get(col, "Unknown"))

        logger.debug("Loaded specification",
                     extra={
                         'transform': self.name,
                         'data': json.dumps(self.transform_spec, indent=4)
                     })

    def validate_tspec(self):
        """
        Validate specification
        """
        tspec = self.transform_spec
        if tspec is None:
            raise Exception("Could not find transform specification")

        if not isinstance(tspec, dict) or len(tspec) == 0:
            raise Exception("Transform specification is either not a dictionary or is empty ")

        if "spec" not in tspec:
            raise Exception("Transform specification required 'spec' element")
        if not isinstance(tspec['spec'], dict) or len(tspec['spec']) == 0:
            raise Exception("Transform specification's 'spec' element should a valid non-empty dict")

        spec = tspec['spec']
        for section in ['credentials', 'input_datasets', 'output_datasets',  'queries']:
            if section in spec:
                if not isinstance(spec[section], list):
                    raise Exception(f"Specification section {section} should be a list of dicts")
                for entry in spec[section]:
                    if not isinstance(entry, dict) or len(entry) == 0:
                        raise Exception(f"Specification section {section} should be a list of non-trivial dicts")

        missing = []
        if 'input_queries' in spec:
            input_queries = spec['input_queries']
            for qspec in input_queries:
                for key  in ['name', 'database', 'filename', 'credential']:
                    if key not in qspec:
                        missing.append(f"[input_queries] {key}")

        if len(missing) > 0:
            logger.error("Missing configuration elements",
                         extra={
                             'transform': self.name,
                             'data': ", ".join(missing)
                         })
            raise Exception("Invalid configuration")

    def load_tspec(self):
        """
        Find and load transform specification

        This looks for spec.json and spec.yaml
        """
        subclspath = sys.modules[self.__module__].__file__
        subclsdir = os.path.dirname(subclspath)
        spec = None

        if exists(join(subclsdir, "spec.json")):
            spec = json.load(open(join(subclsdir, "spec.json")))
        elif  exists(join(subclsdir, "spec.yaml")):
            spec = yaml.safe_load(open(join(subclsdir, "spec.yaml")))
        elif  exists(join(subclsdir, "spec.yml")):
            spec = yaml.safe_load(open(join(subclsdir, "spec.yml")))

        if spec is None:
            raise Exception("Transform specification file (spec.json/spec.yaml) is required")

        return spec

    def load_input_queries(self):
        """
        Find and load input queries at the time of initialization.
        """

        subclspath = sys.modules[self.__module__].__file__
        subclsdir = os.path.dirname(subclspath)

        notfound = []
        queries = self.transform_spec['spec'].get('input_queries',[])
        for q in queries:
            """
            q = {
                "name": "edit_name",
                "description": "edit description",
                "credential": "tdb"
            }
            """
            name = q['name']
            filenames = [
                name,
                f"{name}.sql",
                f"{name}.mql",
            ]

            if "filename" in q:
                filenames.append(q['filename'])

            found = False
            for f in filenames:
                path = os.path.join(subclsdir, f)
                try:
                    if not os.path.exists(path):
                        continue
                    content = open(path).read()
                    if name in self.transform_input_queries:
                        logger.warning(f"Duplicate query: {name}")
                        continue
                    self.transform_input_queries[name] = {
                        'name': name,
                        "filename": f,
                        'template': content
                    }
                    found = True
                    break
                except:
                    logger.exception(f"Failed to read: {name}",
                                     extra={
                                         'transform': self.name
                                     })

            if not found:
                notfound.append(name)

        if len(notfound) > 0:
            logger.info("Input queries not loaded: {len(notfound)}",
                        extra={
                            'transform': self.name,
                            'data': ", ".join(notfound)
                        })

    def preload_clean_args(self, args):
        """
        Load the credentials from the transform specification
        """

        args = super().preload_clean_args(args)

        # Gather credentials
        if 'transform_credentials' in args:
            raise Exception("conflict in args: transform_credentials")
        args['transform_credentials'] = {}

        # Global credentials
        credentials = self.transform_spec.get('spec', {}).get('credentials', [])
        credentials = [c['name'] for c in credentials if ((isinstance(c, dict)) and ('name' in c))]

        # Load credentials specified in input or output queries
        for section in ['input_datasets', 'output_datasets',
                        'input_queries', 'output_queries']:
            for d in self.transform_spec.get('spec', {}).get(section, []):
                if (isinstance(d, dict ) and ('credential' in d)):
                    if d['credential'] not in credentials:
                        credentials.append(d['credential'])

        for name in credentials:
            if name in args['transform_credentials']:
                continue
            args['transform_credentials'][name] = self.get_credentials(name)

        return args


    def get_input_query_params(self, qspec):
        """
        Get the rendering parameters, if any, for the query template

        Args:
           qspec (dict): Specification of the query (filename etc.)

        Returns
           dict: Dictionary with parameters that is passed to the query template renderer

        """
        return {}

    def get_input_query_template(self, qspec):
        """
        Get the input query corresponding to a name
        """

        name = qspec['name']
        if name not in self.transform_input_queries:
            raise Exception(f"Query not found: {name}")

        return self.transform_input_queries[name]['template']

    def get_mongo_client(self, qspec):

        cred = qspec['credential']
        cred = self.args['transform_credentials'][cred]

        # Connect to server
        uri = cred['uri']
        client = pymongo.MongoClient(uri)

        return client

    def execute_mongo_query(self, qspec, query):

        name = qspec['name']

        try:

            # Get the client...
            client = self.get_mongo_client(qspec)

            # Get the database
            db = client[qspec['database']]

            # Now run the pymongo query..
            docs = eval(query)
            if not isinstance(docs, list):
                docs = list(docs)

            df = pd.DataFrame(docs)

        except:
            logger.exception(f"Error while running {name}")
            raise

        return df, {}

    def get_sql_engine(self, qspec):

        cred = qspec['credential']
        cred = self.args['transform_credentials'][cred]

        # Connect to server
        uri = cred['uri']

        # Create an engine
        engine = create_engine(uri)

        return engine

    def execute_sql_query(self, qspec, query):

        name = qspec['name']

        try:

            # Get the connection
            engine = self.get_sql_engine(qspec)

            # Run the query
            df = pd.read_sql(query, con=engine)

        except:
            logger.exception(f"Error while running {name}")
            raise

        return df, {}

    def execute_input_query(self, qspec, query):
        """
        Override this for non-SQL backends
        """

        cred = qspec['credential']
        cred = self.args['transform_credentials'][cred]

        dbtype = cred.get('dbtype', 'unknown')
        if dbtype == 'mongo':
            return self.execute_mongo_query(qspec, query)
        elif dbtype in ['postgres', 'mysql']:
            return self.execute_sql_query(qspec, query)

        # Now run the query...
        df = pd.DataFrame([{'a': 2}])

        return df, {}

    def read_input_query(self, qspec):
        """
        Run the named query rendered using the params
        and using the credential name
        """
        metadata = {}

        try:

            # First get the params...
            params = self.get_input_query_params(qspec)

            # Instantiate the query template
            query_template = self.get_input_query_template(qspec)

            template = Template(query_template)
            query = template.render(**params)

            df, metadata = self.execute_input_query(qspec, query)

        except:
            logger.exception(f"Unable to instantiate the query: {qspec['name']}")
            raise

        return df, {}

    def get_supported_extra_args(self):

        default_extra_args = super().get_supported_extra_args()

        # Now add some more...
        tspec = self.transform_spec
        more_args = tspec.get('spec', {}).get('args', [])

        # Now check if all the required elements exist in the test
        # data...
        tspec = self.transform_spec
        more_args = tspec.get('spec', {}).get('args', [])

        missing_args = {}

        conf = self.testdata.get("conf", {})
        args = conf.get('args', {})

        # Now check if we have all required
        for a in more_args:
            name = a['name']
            default = a['default']
            required = a['required']
            if ((required) and (name not in args)):
                missing_args[name] = default

        if len(missing_args) == 0:
            return default_extra_args + more_args

        msg = "Missing arguments: " + ", ".join(missing_args.keys())  + "\n"
        instance = list(missing_args.items())[0]
        msg += "These arguments can be specified as:\n"
        msg += """\n
def __init__(self, *args, **kwargs):
      super().__init__(*args, *kwargs)
      self.testdata = {
          "conf": {
              "args": {
                  "%s": "%s"
                  ....
              }
          }
      }
        """ % instance
        logger.error("Missing args in testdata",
                     extra={
                         'transform': self.name,
                         'data': msg
                     })
        raise Exception("Missing args in testdata")



class Trigger(Sink):
    """
    This transform is expected to run after the computation
    is over. This is a particular kind of sync

    .. deprecated:: 2.0

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NotificationIntegration(Integration):
    """
    Search interface dashboard

    .. deprecated:: 2.0

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tags.extend(["integration", "notification"])


class DatabaseIntegration(Integration):
    """
    Write to database

    .. deprecated:: 2.0

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tags.extend(["integration", "database"])



# event name definitions
_EVNT_READY = "ready"
_EVNT_DATA = "source-loaded"
_EVNT_INDEXCREATE = "index-created"
_EVNT_INDEXLOAD = "index-loaded"
_EVNT_INDEXSAVE = "index-persisted"
_EVNT_INDEXADD = "index-doc-added"
_EVNT_QUERY = "query"
_EVNT_SANITIZE = "sanitize"
_EVNT_SPEC = "spec-loaded"
_EVNT_DOCGEN = "doc-gen"
_EVNT_STATECLEAR = "state-clear"
_EVNT_STATESTORE = "state-store"
_EVNT_STATERESTORE = "state-restore"



import os
import json
import logging
from enrichsdk.core.node import Node

logger = logging.getLogger("app")


class Schema(object):
    """
    This class allows one to use load exported catalog data
    """

    def __init__(self, filename):

        if not os.path.exists(filename):
            raise Exception("Invalid or missing schema file: {}".format(filename))

        try:
            catalogs = json.load(open(filename))
        except:
            logger.exception("Unable to read filename")
            raise Exception("Unsupported format: {}".format(filename))

        if isinstance(catalogs, dict):
            catalogs = [catalogs]

        for catalog in catalogs:
            if ("schema" not in catalog) or (catalog["schema"] != "Catalog"):
                raise Exception("Only catalog level exports supported")

        self.catalogs = catalogs

    def find_source(self, catalogname, sourcename):

        catalog = None
        for c in self.catalogs:
            if c["name"] == catalogname:
                catalog = c
                break

        if catalog is None:
            return None

        source = None

        for s in catalog["sources"]:
            if s["name"] == sourcename:
                source = s
                break

        return source

    def get_column_docs(self, catalogname, sourcename):
        """
        Summarize the output in form that it usable by the
        transform
        """

        outputs = {}

        source = self.find_source(catalogname, sourcename)
        if source is None:
            # Didnt find the source
            return outputs

        for c in source["columns"]:
            colname = c["name"]
            coldesc = c.get("description", c.get("desc", ""))
            notes = c.get("notes", "")

            outputs[colname] = coldesc + "\n" + notes

        return outputs


class TransformSchemaMixin(object):
    def documentation_from_catalog(self, catalog, frames):

        # Check if the transform is an instance of the sdk
        if not isinstance(self, Node):
            raise Exception("Transform should be an instance of the Node")

        # Check if the catalog exists
        catalog = self.config.get_file(catalog)
        schema = Schema(catalog)
        outputs = self.outputs

        for frame, details in frames.items():
            catalogname = details["catalog"]
            sourcename = details["source"]

            # Initialize
            if frame not in outputs:
                outputs[frame] = {}

            frameoutputs = schema.get_column_docs(catalogname, sourcename)
            if len(frameoutputs) == 0:
                continue

            outputs[frame].update(frameoutputs)



import json
from enum import Enum
from enrichsdk.app.utils import generate_id
import logging

logger= logging.getLogger('app')


class BaseComponent():
    def __init__(self, name, ctype, desc):
        self.name = name
        self.ctype = ctype
        self.description = desc
        self.id = generate_id(self.name)

    def get_component_type(self):
        return self.ctype


    def to_json(self):
        return json.dumps(self.__dict__)


class CustomComponentTypes(str, Enum):
    Notes = 'Notes'
    Headers = 'Headers'
    Body = 'Body'
    Media = 'Media'

    def get_name(self):
        return str(self.value)


class CustomComponent(BaseComponent):
    def __init__(self, name, custom_component_type):
        super().__init__(name, custom_component_type)


""" {   "search_params":
        {
            "persona": persona_name,
            "table": table_name
        },
        "search_query": request.GET.get('query', None),
        "search_column_names": search_columns_names,
        "url": f"?persona={persona_name}&table={table_name}&query="
    } """


class SearchComponent(BaseComponent):
    def __init__(self, name):
        self.ctype = 'Search'
        super().__init__(name, self.ctype)
        self.search_columns_names = []
        self.persona_name = ''
        self.url = ''
        self.table_name = ''
        self.search_query = ''

    def set_search_column_names(self, value):
        self.search_columns_names = value

    def get_search_column_names(self):
        return self.search_columns_names

    @property
    def search_query_attr(self):
        return self.search_query

    @search_query_attr.setter
    def search_query_attr(self, value):
        self.search_query = value

    @property
    def persona_name_attr(self):
        return

    """
    {
        "id": "notes",
        "label": "",
        "description": notes,
        "type": ctype,
        "icon": "",
        "class": "",
        "template": "notes"
    }
    """


class NotesType(str, Enum):
    LIST = 'list'
    PARAGRAPH = 'notes'


class NotesComponent(BaseComponent):
    def __init__(self, name, label, desc, icon, template):
        super().__init__(name, CustomComponentTypes.Notes, desc)
        self.label = label
        self.type = NotesType.LIST
        self.icon = icon
        self.css_class = ''
        self.template = template

    @property
    def css_class_attr(self):
        return self.css_class

    @css_class_attr.setter
    def css_class_attr(self, value):
        self.css_class = value

    # @property
    # def label_attr(self):
    #     return self.css_class
    #
    # @label_attr.setter
    # def label_attr(self, value):
    #     self.label = value
    #
    # @property
    # def desc_attr(self):
    #     return self.desc
    #
    # @desc_attr.setter
    # def desc_attr(self, value):
    #     self.desc = value
    #
    # @property
    # def icon_attr(self):
    #     return self.icon
    #
    # @icon_attr.setter
    # def icon_attr(self, value):
    #     self.icon = value
    #
    # @property
    # def template_attr(self):
    #     return self.template
    #
    # @template_attr.setter
    # def template_attr(self, value):
    #     self.template = value

    def type_attr(self):
        return self.type

    def type_attr(self, value):
        self.type = value



class MediaType(str, Enum):
    IMAGE = 'image'
    VIDEO = 'video'


class MediaComponent(BaseComponent):
    def __init__(self, name):
        super().__init__(name)
        self.label = ''
        self.description = ''
        self.type = MediaType.IMAGE
        self.icon = ''
        self.css_class = ''
        self.template = ''


class HeaderComponent(BaseComponent):
    def __init__(self, name):
        super().__init__(name, CustomComponentTypes.Headers, CustomComponentTypes.Headers.get_name())
        self.components = {}

    def add_component(self, component_type, component):
        if isinstance(component_type, CustomComponentTypes):
            self.components[component_type.get_name()] = component
        else:
            logger.error("Provide correct type of Custom Component")

    def get_components(self):
        return self.components


class BodyComponent(BaseComponent):
    def __init__(self, name):
        super().__init__(name, CustomComponentTypes.Body, CustomComponentTypes.Body.get_name())
        self.components = {}

    def add_component(self, component):

        if component:
            component_type = component.get_component_type()
            component_name = component_type.get_name().lower()
            if isinstance(component_type, CustomComponentTypes):

                if component_name in self.components:
                    self.components[component_name].append(component)

                else:
                    components = [component]
                    self.components[component_name] = components

            else:
                logger.error("Invalid component type")
        else:
            logger.error("Component can't be null")

    def get_components(self):
        return self.components



from django.urls import reverse
from enrichsdk.lib.customer import find_usecase

spec = {
    'name': 'APPNAMEApp',
    'description': "Custom application",
    'usecase': find_usecase(__file__),
    'icon': 'line-graph_benchmarking'
}

def get_spec():
    return spec



import os
import re
import json
import traceback
import logging
import seaborn as sns
import matplotlib.pyplot as plt
from io import BytesIO, StringIO
import base64

from django.http import QueryDict
from django.shortcuts import render
from django.http import HttpResponse, JsonResponse
from django.conf import settings
from django.db import models
from django.contrib.auth.decorators import login_required
from django.views.decorators.http import require_http_methods
from django.http import HttpResponseRedirect, Http404
from django.contrib import messages
from django.urls import reverse, resolve
from django.conf.urls import url, include

import s3fs
import gcsfs

from users.decorators import log_activity
from dashboard.lib import log_app_activity

from enrichsdk.utils import SafeEncoder
from enrichapp.spec import validate_spec as default_validate_spec, fill_action_gaps
from enrichsdk.lib import get_credentials_by_name
from enrichsdk.utils import SafeEncoder
from enrichsdk.datasets import Doodle
from enrichsdk.lib.integration import send_html_email
from enrichsdk.lib import get_default_from_email

from email import encoders
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

import pandas as pd
from pretty_html_table import build_table

logger= logging.getLogger('app')


###############################################
# Main class
###############################################
class AppBase():
    """
    A shareable view across apps
    """

    def __init__(self):
        self.name = "outliers"
        self.verbose_name = "Outliers"
        self.category = "outliers"
        self.templates = {
            'policy_index': f'enrichapp/policyapp/policy_index.html',
            'policy_add': f'enrichapp/policyapp/policy_add.html',
            'policy_select_dataset': f'enrichapp/policyapp/policy_select_dataset.html',
            'policy_select_columns': f'enrichapp/policyapp/policy_select_columns.html',
            'policy_select_columns_helper': f'enrichapp/policyapp/policy_select_columns_helper.html',
            'result_index': f'enrichapp/policyapp/result_index.html',
            'result_detail': f'enrichapp/policyapp/result_detail.html',
            'workflow_index': f'enrichapp/policyapp/workflow_index.html',

        }
        self.select_columns_post_spec = {
            'list_columns': []
        }

    def __str__(self):
        return f"[{self.name}] {self.verbose_name}"

    def get_readme(self):
        return """This is a generic policy app"""

    def get_template(self, spec, name):
        return spec.get('templates',{}).get(name, self.templates[name])

    def get_model(self, spec, name):

        from enrichapp.dashboard.globalapp import lib as globallib

        if (('models' in spec) and (name in spec['models'])):
            return spec['models'][name]

        return globallib.get_model(name)

    def get_form(self, spec, name):

        from enrichapp.dashboard.globalapp import lib as globallib

        if (('forms' in spec) and (name in spec['forms'])):
            return spec['forms'][name]

        return globallib.get_form(name)

    def validate_spec(self, spec):

        fill_action_gaps(spec, context={ })

        errors = default_validate_spec(spec)
        if len(errors) > 0:
            return len(errors) == 0, errors

        if 'namespace' not in spec:
            errors.append(f"namespace should be specified")

        if (('doodle' not in spec) or
            ((not isinstance(spec['doodle'], dict)) and (not callable(spec['doodle'])))):
            errors.append(f"Doodle metadata server is not specified")

        if ('data' not in spec):
            errors.append(f"Data section must be mentioned")
        else:
            missing = [col for col in ['type', 'cred', 'root'] if col not in spec['data']]
            if len(missing) > 0:
                errors.append(f"Missing data elements: {','.join(missing)}")

        return len(errors) == 0, errors

    def check_prerequisites(self, request, spec):

        valid, errors = self.validate_spec(spec)
        if not valid:
            error = "Invalid specification"
            messages.error(request, error)
            msg = "Errors: " + json.dumps(errors, indent=4, cls=SafeEncoder)
            msg += "Spec:\n" + json.dumps(spec,  indent=4, cls=SafeEncoder)

            logger.error(error,
                         extra={
                             'transform': 'Dashboard',
                             'data': msg
                         })
            return valid, HttpResponseRedirect(reverse('dashboard:application_index'))

        return valid, None

    def get_urlpatterns(self):
        urlpatterns = [
            url('^[/]?$', self.index, name="index"),
            url('^configurations[/]?$', self.policy_index, name="policy_index"),
            url('^configurations/add[/]?$', self.policy_add, name="policy_add"),
            url('^configurations/(?P<policy_id>[0-9]+)/toggle[/]?$', self.policy_toggle, name="policy_toggle"),
            url('^configurations/select/dataset[/]?$', self.policy_select_dataset, name="policy_select_dataset"),
            url('^configurations/select/columns[/]?$', self.policy_select_columns, name="policy_select_columns"),
            url('^results[/]?$', self.result_index, name="result_index"),
            url('^results/detail[/]?$', self.result_detail, name="result_detail"),
            url('^workflows/action[/]?$', self.workflow_action, name="workflow_action"),
            url('^workflows[/]?$', self.workflow_index, name="workflow_index")
        ]

        return urlpatterns

    @log_activity("app_index", nature="application")
    def index(self, request, spec):

        r = resolve(request.path)

        valid, redirect = self.check_prerequisites(request, spec)
        if not valid:
            return redirect

        return HttpResponseRedirect(reverse(r.namespace + ":policy_index"))

    @log_activity("app_policy_index", nature="application")
    def policy_index(self, request, spec):
        """
        Policy index...
        """

        r = resolve(request.path)

        valid, redirect = self.check_prerequisites(request, spec)
        if not valid:
            return redirect

        usecase = spec['usecase']

        template = self.get_template(spec, 'policy_index')

        namespace = spec.get('namespace', 'default')
        try:
            namepace = spec['namespace']
            PolicyModel    = self.get_model(spec, 'app_policy')
            policies = PolicyModel.objects\
                                  .filter(appname=self.name, namespace=namespace)\
                                  .order_by("-id")

        except:
            logger.exception(f"Unable to get {self.name} policies. Please see server log")
            #return HttpResponseRedirect(reverse(r.namespace + ":index"))

        return render(request,
                      template,
                      {
                          'app': self,
                          'spec': spec,
                          'usecase': usecase,
                          'basenamespace': r.namespace,
                          'policies': policies
                      })


    @log_activity("app_policy_toggle", nature="application")
    def policy_toggle(self, request, spec, policy_id):
        """
        Activate/Deactivate policy
        """

        r = resolve(request.path)

        valid, redirect = self.check_prerequisites(request, spec)
        if not valid:
            return redirect

        usecase = spec['usecase']


        namespace      = spec.get('namespace', 'default')
        PolicyModel    = self.get_model(spec, 'app_policy')
        policies       = PolicyModel.objects.filter(pk=policy_id, appname=self.name, namespace=namespace)
        if policies.count() == 0:
            messages.error(request, "No policy found")
            return HttpResponseRedirect(reverse(r.namespace + ":policy_index"))

        policy = policies[0]
        policy.active = not policy.active
        policy.save()
        messages.success(request, "Policy updated")
        return HttpResponseRedirect(reverse(r.namespace + ":policy_index"))

    @log_activity("app_policy_select_dataset", nature="application")
    def policy_select_dataset(self, request, spec):
        """
        Select datasets to process
        """
        r = resolve(request.path)

        valid, redirect = self.check_prerequisites(request, spec)
        if not valid:
            return redirect

        usecase = spec['usecase']

        template = self.get_template(spec, 'policy_select_dataset')
        cred = spec['doodle']
        if callable(cred):
            cred = cred(spec)

        doodle = Doodle(cred)

        # Collect the sources
        sources = doodle.list_sources()

        return render(request,
                      template,
                      {
                          'app': self,
                          'spec': spec,
                          'usecase': usecase,
                          'basenamespace': r.namespace,
                          'sources': sources
                      })

    def validate_policy_config(self, request, spec, config):
        """
        Validate the config generated by the backend
        """
        return len(config) > 0

    def policy_generate_config(self, request, spec):
        """
        Process the select columns form
        """

        r = resolve(request.path)

        valid, redirect = self.check_prerequisites(request, spec)
        if not valid:
            return redirect

        source_id = request.GET.get('source_id', None)
        if ((source_id is None) or (request.method != "POST")):
            logger.error("Missing source id or not a post")
            return HttpResponseRedirect(reverse(r.namespace + ":policy_select_dataset"))

        config = request.GET.dict()
        config.update(request.POST.dict())
        listcols = self.select_columns_post_spec['list_columns']
        for col in listcols:
            config[col] = request.POST.getlist(col)

        # Cleanup
        config.pop('csrfmiddlewaretoken',None)

        valid = self.validate_policy_config(request, spec, config)

        return valid, config

    @log_activity("app_policy_select_columns", nature="application")
    def policy_select_columns(self, request, spec):
        """
        Select columns
        """
        r = resolve(request.path)

        valid, redirect = self.check_prerequisites(request, spec)
        if not valid:
            return redirect

        usecase = spec['usecase']
        template = self.get_template(spec, 'policy_select_columns')

        # Now access the metadata server..
        source_id = request.GET.get('source_id', None)
        if source_id is None:
            logger.error("Missing source id")
            return HttpResponseRedirect(reverse(r.namespace + ":policy_select_dataset"))
        source_version = request.GET.get('source_version', 'v1')

        # Handle the post from selection...
        if request.method == "POST":
            valid, config = self.policy_generate_config(request, spec)
            if not valid:
                # Errors got incorporated in the validate function..
                messages.error(request, "Invalid selection")
                return HttpResponseRedirect(reverse(r.namespace + ":policy_select_columns") + f"?source_id={source_id}&source_version={source_version}")
            try:
                config = json.dumps(config)
                return HttpResponseRedirect(reverse(r.namespace + ":policy_add") + f"?config={config}")
            except Exception as e:
                logger.exception("Error while redirecting to policy add")
                messages.error(request, "Error while adding a policy")
                return HttpResponseRedirect(reverse(r.namespace + ":policy_select_columns") + f"?source_id={source_id}&source_version={source_version}")


        # Collect the sources
        cred = spec['doodle']
        if callable(cred):
            cred = cred(spec)
        doodle = Doodle(cred)

        source = doodle.get_source(source_id)
        features = doodle.list_features(source_id=source['id'])

        # Now render
        return render(request,
                      template,
                      {
                          'app': self,
                          'spec': spec,
                          'usecase': usecase,
                          'basenamespace': r.namespace,
                          'select_columns_helper': self.get_template(spec, 'policy_select_columns_helper'),
                          'source': source,
                          'columns': features
                  })


    @log_activity("app_policy_add", nature="application")
    def policy_add(self, request, spec):
        """
        Add a policy
        """

        r = resolve(request.path)

        valid, redirect = self.check_prerequisites(request, spec)
        if not valid:
            return redirect

        usecase = spec['usecase']
        template = self.get_template(spec, 'policy_add')

        # => Get models
        namespace   = spec.get('namespace','default')
        PolicyModel = self.get_model(spec, 'app_policy')
        PolicyForm  = self.get_form(spec,  'app_policy')

        config = request.GET.get('config', '{}')
        config = json.loads(config)

        if request.method != "POST":

            initial = {
                "config": config,
            }

            form = PolicyForm(initial=initial)

            return render(request,
                          template,
                          {
                              'app': self,
                              'spec': spec,
                              'usecase': usecase,
                              'basenamespace': r.namespace,
                              'form': form
                          })


        form = PolicyForm(request.POST)
        if form.is_valid():
            policy = form.save(commit=False)
            policy.appname = self.name
            policy.namespace = namespace
            policy.created_by = request.user
            policy.modified_by = request.user
            policy.active = True
            policy.save()

            log_app_activity(request, self.name, spec, "policy added", policy.desc)

            messages.success(request, 'Policy {} saved!'.format(policy.name))
            return HttpResponseRedirect(reverse(r.namespace + ":policy_index"))


        return render(request,
                      template,
                      {
                          'app': self,
                          'spec': spec,
                          'usecase': usecase,
                          'basenamespace': r.namespace,
                          'form': form
                      })

    ####################################################3
    # Result
    ####################################################3
    def get_fs_handle(self, cred):
        """
        Get s3/gcs filesystem handle..
        """
        cred = get_credentials_by_name(cred)
        nature = cred['nature']
        if nature not in ['s3', 'gcs']:
            raise Exception(f"Unknown credentials: {nature}")

        if nature == 's3':
            config_kwargs={
                'signature_version': 's3v4'
            }

            if 'region' in cred:
                config_kwargs['region_name'] = cred['region']

            fshandle = s3fs.S3FileSystem(
                key    = cred['access_key'],
                secret = cred['secret_key'],
                config_kwargs=config_kwargs,
                use_listings_cache=False
            )
        else:
            fshandle = gcsfs.GCSFileSystem(
                token=cred['keyfile']
            )

        return fshandle

    def get_result_list(self, request, spec, policies):

        try:
            available = {}

            data = spec['data']
            handle = self.get_fs_handle(data['cred'])

            appname = self.name
            namespace = spec['namespace']

            root = os.path.join(data['root'], appname, namespace)

            fallback = spec.get('fallback',False)
            if fallback:
                if not handle.exists(root):
                    messages.warning(request, "Result root directory is missing. choosing alternative")
                    namespaces = handle.ls(os.path.join(data['root'], appname))
                    if len(namespaces) == 0:
                        return []
                    root = namespaces[0]


            # What should I be looking at?
            relevant_configurations = [p.name for p in policies if ((p.appname == self.name) &
                                                                    (p.namespace == namespace))]

            # available paths
            # Look for the subset within the namespace
            paths = handle.find(root, maxdepth=2, withdirs=True)
            confmap = {}
            for p in paths:
                p = p.replace(root+ "/", "").split("/")
                if len(p) != 2:
                    continue
                confname, dt = p
                if confname not in confmap:
                    confmap[confname] = {
                        'dates': []
                    }
                confmap[confname]['dates'].append(dt)

            for confname in list(confmap.keys()):
                confmap[confname]['latest'] = max(confmap[confname]['dates'])

            available = {c: detail for c, detail in confmap.items() if ((c in relevant_configurations) or True)}

            if hasattr(handle, 'close'):
                handle.close()

        except:
            #traceback.print_exc()
            logger.exception(f"Error in accessing {spec['name']} results")

        return available

    def get_one_result_path(self, request, spec, policy, handle, path):

        # Read the s3/gcs file. Check if this is a string
        content = json.load(handle.open(path))
        if isinstance(content, str):
            content = json.loads(content)

        return content

    def get_result(self, request, spec, policy):
        """
        Get data from s3 and other places
        """
        try:

            data = spec['data']
            handle = self.get_fs_handle(data['cred'])

            appname = self.name
            namespace = spec['namespace']

            root = os.path.join(data['root'], appname, namespace)

            # may be different from policy.name
            name = request.GET.get('name')

            fallback = spec.get('fallback',False)
            if fallback:
                if not handle.exists(root):
                    messages.warning(request, "Result root directory is missing. choosing alternative")
                    namespaces = handle.ls(os.path.join(data['root'], appname))
                    root = namespaces[0]

            names = [os.path.basename(n) for n in handle.ls(root)]
            if name not in names:
                raise Exception(f"Could not find {name}")

            dates = sorted(handle.ls(os.path.join(root, name)), reverse=True)
            if len(dates) == 0:
                return None, None

            paths = sorted(handle.ls(dates[0]), reverse=True)
            if len(paths) == 0:
                return None

            content = None
            paths = [p for p in paths if not p.endswith('metadata.json')]
            for p in paths:
                try:
                    content = self.get_one_result_path(request,
                                                       spec, policy,
                                                       handle, p)
                    return p, content
                except:
                    #traceback.print_exc()
                    continue

            if hasattr(handle, 'close'):
                handle.close()

        except:
            logger.exception(f"Error in accessing {policy.name} results")

        return None, None

    @log_activity("app_result_index", nature="application")
    def result_index(self, request, spec):
        """
        Show the available results
        """

        r = resolve(request.path)

        valid, redirect = self.check_prerequisites(request, spec)
        if not valid:
            return redirect

        usecase = spec['usecase']

        template = self.get_template(spec, 'result_index')

        namespace = spec.get('namespace', 'default')
        try:
            namepace = spec['namespace']
            PolicyModel    = self.get_model(spec, 'app_policy')
            policies = PolicyModel.objects\
                                  .filter(appname=self.name,
                                          namespace=namespace)\
                                  .order_by("-id")
        except Exception as e:
            messages.error(request, f"Error! {e}")
            logger.exception(f"Unable to get {namespace} configurations. Please see server log")
            return HttpResponseRedirect(reverse(r.namespace + ":policy_index"))

        available = self.get_result_list(request, spec, policies)

        return render(request,
                      template,
                      {
                          'app': self,
                          'spec': spec,
                          'usecase': usecase,
                          'basenamespace': r.namespace,
                          'policies': policies,
                          'available': available
                      })

    @log_activity("app_result_detail", nature="application")
    def result_detail(self, request, spec):
        """
        Show the available results
        """
        from dashboard.models import AppActionTarget

        r = resolve(request.path)

        valid, redirect = self.check_prerequisites(request, spec)
        if not valid:
            return redirect

        fallback = spec.get('fallback', False)
        usecase = spec['usecase']

        template = self.get_template(spec, 'result_detail')

        name = request.GET.get('name', None)
        if name is None:
            messages.error(request, "Missing configuration")
            return HttpResponseRedirect(reverse(r.namespace + ":result_index"))

        try:
            namespace      = spec['namespace']
            PolicyModel    = self.get_model(spec, 'app_policy')
            policy         = PolicyModel.objects\
                                          .get(appname=self.name,
                                               namespace=namespace,
                                               name=name,
                                               active=True
                                          )
        except Exception as e:
            logger.exception(f"Unable to get configuration for {name}")
            if not fallback:
                messages.error(request, f"Unable to get configuration for {name}")
                return HttpResponseRedirect(reverse(r.namespace + ":result_index"))

            messages.warning(request, f"Unable to get configuration for {name}")
            try:
                policy = PolicyModel.objects.filter(appname=self.name,namespace=namespace, active=True).first()
            except:
                messages.warning(request, f"No policy defined for fallback")
                return HttpResponseRedirect(reverse(r.namespace + ":result_index"))

        path, data = self.get_result(request, spec, policy)
        if data is None:
            messages.error(request, f"Unable to get data for configuration {name}")
            logger.exception(f"Unable to get data for configuration {name}")
            return HttpResponseRedirect(reverse(r.namespace + ":result_index"))

        # Workflow targets
        supported_actions = self.get_supported_actions(request, spec, policy)
        targets = AppActionTarget.objects.filter(nature__in=supported_actions)

        return render(request,
                      template,
                      {
                          'app': self,
                          'spec': spec,
                          'usecase': usecase,
                          'basenamespace': r.namespace,
                          'name': name,
                          'policy': policy,
                          'data': data,
                          'path': path,
                          'targets': targets
                      })

    ################################################################
    # Workflow Management
    ################################################################
    def get_supported_actions(self, request, spec, policy):
        return ['Email']

    def take_action(self, request, spec, name, data, target):

        r = resolve(request.path)
        detailurl = request.build_absolute_uri(reverse(r.namespace + ":result_detail")) + f"?name={name}"

        try:
            if target.nature == "Email":

                """{
                    "emails": [
                       "pingali@scribbledata.io"
                     ]
                }"""

                policy = target.policy
                if (('emails' not in policy) and (not isinstance(target.policy['emails'], list))):
                    return JsonResponse({
                        'status': 'failure',
                        'message': "Email service not configured correct. Please contact the admin"
                    })

                user = request.user
                fullname = f"{user.first_name} {user.last_name}".strip()
                if len(fullname) == 0:
                    fullname = user.username

                sender = fullname + "<" + get_default_from_email() +">"
                receivers = policy['emails']
                subject = f"{spec['name']} - Sharing Observation"
                content = f"<p>{data['message']}</p>"
                content += f"<p>You can find the full details at <a href='{detailurl}'>{spec['name']}</a> page. Here is quick context:</p>"
                table = json.loads(data['data'])
                s = pd.Series(table).reset_index()
                s.columns = ['Attribute', 'Value']
                content += build_table(s,
                                       'blue_light',
                                       font_size='small',
                                       font_family='Roboto, sans-serif')

                msg = MIMEMultipart()
                msg.attach(MIMEText(content, "html"))
                send_html_email(content=content,
                                sender=sender,
                                receivers=receivers,
                                subject=subject)

                return JsonResponse({
                    'status': 'success',
                    'message': "Email sent successfully"
                })

            return JsonResponse({
                'status': "failure",
                "message": f"Workflow service not configured correctly. Please contact admin"
            })

        except Exception as e:
            logger.exception("Unable to send email")
            return JsonResponse({
                'status': "failure",
                "message": f"{target.nature} message action could not be taken. {str(e)}"
            })

    @log_activity("workflow_action", nature="application")
    def workflow_action(self, request, spec):
        """
        Show the workflows for the given data
        """

        from dashboard.models import AppActionTarget

        r = resolve(request.path)

        valid, redirect = self.check_prerequisites(request, spec)
        if not valid:
            return redirect

        usecase = spec['usecase']

        name = request.GET.get('name', None)
        if name is None:
            logger.exception("Missing name of the result")
            return HttpResponseRedirect(reverse(r.namespace + ":result_index"))

        data = request.POST.dict()

        try:
            try:
                target = AppActionTarget.objects.get(pk=data['actionid'])
            except:
                raise Exception("Internal error. Missing or invalid action id")

            messages.success(request, "Processed the post")
            status = "success"

            # Make this happen...
            status = self.take_action(request, spec, name, data, target)

        except Exception as e:
            return JsonResponse({
                'status': 'failure',
                'message': f"Failed to trigger action. {str(e)}"
            })

        return status

    @log_activity("workflow_index", nature="application")
    def workflow_index(self, request, spec):
        """
        Show the workflows for the given data
        """

        from dashboard.models import AppActionTarget


        r = resolve(request.path)

        valid, redirect = self.check_prerequisites(request, spec)
        if not valid:
            return redirect

        usecase = spec['usecase']

        name = request.GET.get('name', None)
        if name is None:
            logger.exception("Missing name of the result")
            return HttpResponseRedirect(reverse(r.namespace + ":result_index"))

        data = request.GET.get('data', '{}')
        try:
            data = json.loads(data)
        except:
            logger.exception("Missing data dictionary required for workflows")
            return HttpResponseRedirect(reverse(r.namespace + ":result_index"))

        template = self.get_template(spec, 'workflow_index')

        targets = AppActionTarget.objects.filter(nature__in=self.get_supported_actions(request, spec, policy))

        return render(request,
                      template,
                      {
                          'app': self,
                          'spec': spec,
                          'usecase': usecase,
                          'basenamespace': r.namespace,
                          "name": name,
                          'data': data,
                          'targets': targets
                      })

    ################################################################
    # API
    ################################################################
    @log_activity("app_policy_index_api", nature="application")
    def api_policy_index(self, request, instancename):
        """
        Return the queryset
        """
        from dashboard.apiv2 import lookup_specification

        spec = lookup_specification(name=instancename,
                                    category=self.category)
        if spec is None:
            return {
                "status": "failure",
                "message": f"Unknown specification: {self.name}"
            }

        namespace = spec.get('namespace', 'default')
        PolicyModel    = self.get_model(spec, 'app_policy')

        qs = PolicyModel.objects.filter(active=True,
                                        appname=self.name,
                                        namespace=namespace)
        policies = [p.export(spec) for p in qs]

        return {
            "status": "success",
            "data": policies
        }


    ################################################################
    # Ninja Routing/URLS setup
    ################################################################
    def get_router(self):

        from ninja import Router, Query, Schema

        router = Router()

        @router.get("/app/" + self.name + "/{appname}/policies",
                    url_name=self.name+"_policy_index",
                    summary="List policies defined in this app",
                    tags=[self.name])
        def api_policies_index(request, appname):
            """
            List policies
            """
            return self.api_policy_index(request, instancename=appname)

        return router
